from abc import ABC, abstractmethod
from functools import cached_property

import numpy as np

from ...utils.decorators import read_only_array
from ..graph import Graph


class StructuredQuadLayout(ABC):
    @staticmethod
    def corner_nodes(shape):
        n_rows, n_cols = shape
        return (n_rows * n_cols - 1, (n_rows - 1) * n_cols, 0, n_cols - 1)

    @staticmethod
    @abstractmethod
    def links_at_patch(shape):
        ...

    @staticmethod
    @abstractmethod
    def nodes_at_link(shape):
        ...

    @staticmethod
    @abstractmethod
    def horizontal_links(shape):
        ...

    @staticmethod
    @abstractmethod
    def vertical_links(shape):
        ...

    @staticmethod
    @abstractmethod
    def perimeter_nodes(shape):
        ...

    @staticmethod
    @abstractmethod
    def links_at_node(shape):
        ...

    @staticmethod
    @abstractmethod
    def patches_at_link(shape):
        ...

    @staticmethod
    @abstractmethod
    def link_dirs_at_node(shape):
        ...

    @staticmethod
    @abstractmethod
    def patches_at_node(shape):
        ...


class StructuredQuadLayoutCython(StructuredQuadLayout):
    @staticmethod
    def links_at_patch(shape):
        """Get links that define patches for a raster grid.

        Examples
        --------
        >>> from landlab.graph.structured_quad.structured_quad import StructuredQuadLayoutCython
        >>> StructuredQuadLayoutCython.links_at_patch((3, 4))
        array([[ 4,  7,  3,  0], [ 5,  8,  4,  1], [ 6,  9,  5,  2],
               [11, 14, 10,  7], [12, 15, 11,  8], [13, 16, 12,  9]])
        """
        from .ext.at_patch import fill_links_at_patch

        n_patches = (shape[0] - 1) * (shape[1] - 1)
        links_at_patch = np.empty((n_patches, 4), dtype=int)
        fill_links_at_patch(shape, links_at_patch)
        return links_at_patch

    @staticmethod
    def nodes_at_link(shape):
        """
        Examples
        --------
        >>> from landlab.graph.structured_quad.structured_quad import StructuredQuadLayoutCython
        >>> StructuredQuadLayoutCython.nodes_at_link((3, 4))
        array([[ 0,  1], [ 1,  2], [ 2,  3],
               [ 0,  4], [ 1,  5], [ 2,  6], [ 3,  7],
               [ 4,  5], [ 5,  6], [ 6,  7],
               [ 4,  8], [ 5,  9], [ 6, 10], [ 7, 11],
               [ 8,  9], [ 9, 10], [10, 11]])
        """
        from .ext.at_link import fill_nodes_at_link

        n_links = shape[0] * (shape[1] - 1) + (shape[0] - 1) * shape[1]
        nodes_at_link = np.empty((n_links, 2), dtype=int)
        fill_nodes_at_link(shape, nodes_at_link)

        return nodes_at_link

    @staticmethod
    def horizontal_links(shape):
        from .ext.at_link import fill_horizontal_links

        n_horizontal_links = shape[0] * (shape[1] - 1)
        horizontal_links = np.empty(n_horizontal_links, dtype=int)
        fill_horizontal_links(shape, horizontal_links)

        return horizontal_links

    @staticmethod
    def vertical_links(shape):
        from .ext.at_link import fill_vertical_links

        n_vertical_links = (shape[0] - 1) * shape[1]
        vertical_links = np.empty(n_vertical_links, dtype=int)
        fill_vertical_links(shape, vertical_links)

        return vertical_links

    @staticmethod
    def perimeter_nodes(shape):
        from .ext.at_node import fill_perimeter_nodes

        n_perimeter_nodes = 2 * shape[0] + 2 * (shape[1] - 2)
        perimeter_nodes = np.empty(n_perimeter_nodes, dtype=int)
        fill_perimeter_nodes(shape, perimeter_nodes)

        return perimeter_nodes

    @staticmethod
    def links_at_node(shape):
        from .ext.at_node import fill_links_at_node

        n_nodes = shape[0] * shape[1]
        links_at_node = np.empty((n_nodes, 4), dtype=int)
        fill_links_at_node(shape, links_at_node)

        return links_at_node

    @staticmethod
    def patches_at_link(shape):
        from .ext.at_link import fill_patches_at_link

        n_links = shape[0] * (shape[1] - 1) + (shape[0] - 1) * shape[1]
        patches_at_link = np.empty((n_links, 2), dtype=int)
        fill_patches_at_link(shape, patches_at_link)

        return patches_at_link

    @staticmethod
    def link_dirs_at_node(shape):
        from .ext.at_node import fill_link_dirs_at_node

        n_nodes = shape[0] * shape[1]
        link_dirs_at_node = np.empty((n_nodes, 4), dtype=np.int8)
        fill_link_dirs_at_node(shape, link_dirs_at_node)
        return link_dirs_at_node

    @staticmethod
    def patches_at_node(shape):
        from .ext.at_node import fill_patches_at_node

        n_nodes = shape[0] * shape[1]
        patches_at_node = np.empty((n_nodes, 4), dtype=int)
        fill_patches_at_node(shape, patches_at_node)

        return patches_at_node


class StructuredQuadLayoutPython(StructuredQuadLayout):
    @staticmethod
    def links_at_patch(shape):
        n_rows, n_cols = shape
        n_patches = (shape[0] - 1) * (shape[1] - 1)
        links_at_patch = np.empty((4, n_patches), dtype=int)

        patches = np.arange(n_patches, dtype=int).reshape((n_rows - 1, n_cols - 1))
        south_links = patches + np.arange(n_rows - 1).reshape((n_rows - 1, 1)) * n_cols
        links_at_patch[3, :] = south_links.flat
        links_at_patch[2, :] = links_at_patch[3, :] + n_cols - 1
        links_at_patch[1, :] = links_at_patch[3, :] + 2 * n_cols - 1
        links_at_patch[0, :] = links_at_patch[2, :] + 1

        return links_at_patch.T

    @staticmethod
    def nodes_at_link(shape):
        n_rows, n_cols = shape

        nodes_at_link = np.empty(
            (2, n_rows * (n_cols - 1) + (n_rows - 1) * n_cols), dtype=int
        )
        nodes = np.arange(n_rows * n_cols, dtype=int).reshape((n_rows, n_cols))

        nodes_at_link[0, -(n_cols - 1) :] = nodes[-1, :-1]
        nodes_at_link[1, -(n_cols - 1) :] = nodes[-1, 1:]

        head_nodes = nodes_at_link[0, : -(n_cols - 1)].reshape(
            (n_rows - 1, 2 * n_cols - 1)
        )
        head_nodes[:, : n_cols - 1] = nodes[:-1, :-1]
        head_nodes[:, n_cols - 1 :] = nodes[:-1, :]

        tail_nodes = nodes_at_link[1, : -(n_cols - 1)].reshape(
            (n_rows - 1, 2 * n_cols - 1)
        )
        tail_nodes[:, : n_cols - 1] = nodes[:-1, 1:]
        tail_nodes[:, n_cols - 1 :] = nodes[1:, :]

        return nodes_at_link.T

    @staticmethod
    def horizontal_links(shape):
        n_rows, n_cols = shape
        horizontal_links = np.empty((n_rows, n_cols - 1), dtype=int)
        horizontal_links[:, :] = np.arange(n_cols - 1)
        horizontal_links[:, :] += np.arange(n_rows).reshape((n_rows, 1)) * (
            2 * n_cols - 1
        )

        return horizontal_links.reshape(-1)

    @staticmethod
    def vertical_links(shape):
        n_rows, n_cols = shape

        vertical_links = np.empty((n_rows - 1, n_cols), dtype=int)
        vertical_links[:, :] = np.arange(n_cols) + n_cols - 1
        vertical_links[:, :] += np.arange(n_rows - 1).reshape((n_rows - 1, 1)) * (
            2 * n_cols - 1
        )

        return vertical_links.reshape(-1)

    @staticmethod
    def perimeter_nodes(shape):
        n_rows, n_cols = shape
        (
            northeast,
            northwest,
            southwest,
            southeast,
        ) = StructuredQuadLayout.corner_nodes(shape)

        return np.concatenate(
            (
                np.arange(southeast, northeast, n_cols),
                np.arange(northeast, northwest, -1),
                np.arange(northwest, southwest, -n_cols),
                np.arange(southwest, southeast, 1),
            )
        )

    @staticmethod
    def links_at_node(shape):
        n_rows, n_cols = shape

        links_at_node = np.empty((n_rows * n_cols, 4), dtype=int)

        east_links_at_node = links_at_node[:, 0].reshape((n_rows, n_cols))[:, :-1]
        east_links_at_node[:] = StructuredQuadLayoutPython.horizontal_links(
            shape
        ).reshape((n_rows, n_cols - 1))
        west_links_at_node = links_at_node[:, 2].reshape((n_rows, n_cols))[:, 1:]
        west_links_at_node[:] = StructuredQuadLayoutPython.horizontal_links(
            shape
        ).reshape((n_rows, n_cols - 1))

        north_links_at_node = links_at_node[:, 1].reshape((n_rows, n_cols))[:-1, :]
        north_links_at_node[:] = StructuredQuadLayoutPython.vertical_links(
            shape
        ).reshape((n_rows - 1, n_cols))
        south_links_at_node = links_at_node[:, 3].reshape((n_rows, n_cols))[1:, :]
        south_links_at_node[:] = StructuredQuadLayoutPython.vertical_links(
            shape
        ).reshape((n_rows - 1, n_cols))

        (
            northeast,
            northwest,
            southwest,
            southeast,
        ) = StructuredQuadLayout.corner_nodes(shape)
        bottom = slice(southwest, southeast + 1)
        top = slice(northwest, northeast + 1)
        left = slice(southwest, northwest + n_cols, n_cols)
        right = slice(southeast, northeast + n_cols, n_cols)

        for col, edge in enumerate((right, top, left, bottom)):
            links_at_node[edge, col] = -1

        return links_at_node

    @staticmethod
    def patches_at_link(shape):
        n_rows, n_cols = shape
        n_links = shape[0] * (shape[1] - 1) + (shape[0] - 1) * shape[1]
        n_patches = (n_rows - 1) * (n_cols - 1)
        patches = np.arange(n_patches, dtype=int).reshape((n_rows - 1, n_cols - 1))

        patches_at_link = np.empty((2, n_links), dtype=int)
        patches_at_link[0, : n_cols - 1] = -1
        patches_at_link[1, -(n_cols - 1) :] = -1

        right = (0, slice(n_cols - 1, n_links))
        left = (1, slice(0, -(n_cols - 1)))
        for edge in (right, left):
            edge_patches = patches_at_link[edge].reshape((n_rows - 1, 2 * n_cols - 1))
            edge_patches[:, : n_cols - 1] = patches
            edge_patches[:, n_cols - 1] = -1
            edge_patches[:, -(n_cols - 1) :] = patches

        return patches_at_link.T

    @staticmethod
    def link_dirs_at_node(shape):
        n_rows, n_cols = shape
        (
            northeast,
            northwest,
            southwest,
            southeast,
        ) = StructuredQuadLayout.corner_nodes(shape)

        link_dirs_at_node = np.empty((n_rows * n_cols, 4), dtype=np.int8)
        link_dirs_at_node[:, 0] = -1
        link_dirs_at_node[:, 1] = -1
        link_dirs_at_node[:, 2] = 1
        link_dirs_at_node[:, 3] = 1

        bottom = slice(southwest, southeast + 1)
        top = slice(northwest, northeast + 1)
        left = slice(southwest, northwest + n_cols, n_cols)
        right = slice(southeast, northeast + n_cols, n_cols)

        for col, edge in enumerate((right, top, left, bottom)):
            link_dirs_at_node[edge, col] = 0

        return link_dirs_at_node

    @staticmethod
    def patches_at_node(shape):
        n_rows, n_cols = shape

        patches_at_node = np.empty((4, n_rows * n_cols), dtype=int)

        ne = (slice(n_rows - 1), slice(n_cols - 1))
        nw = (slice(n_rows - 1), slice(1, n_cols))
        sw = (slice(1, n_rows), slice(1, n_cols))
        se = (slice(1, n_rows), slice(n_cols - 1))

        patches = np.arange((n_rows - 1) * (n_cols - 1), dtype=int).reshape(
            (n_rows - 1, n_cols - 1)
        )
        for col, nodes in enumerate((ne, nw, sw, se)):
            patches_at_node[col].reshape((n_rows, n_cols))[nodes] = patches

        (
            northeast,
            northwest,
            southwest,
            southeast,
        ) = StructuredQuadLayout.corner_nodes(shape)
        bottom = slice(southwest, southeast + 1)
        top = slice(northwest, northeast + 1)
        left = slice(southwest, northwest + n_cols, n_cols)
        right = slice(southeast, northeast + n_cols, n_cols)

        patches_at_node[0, right] = -1
        patches_at_node[0, top] = -1
        patches_at_node[1, left] = -1
        patches_at_node[1, top] = -1
        patches_at_node[2, left] = -1
        patches_at_node[2, bottom] = -1
        patches_at_node[3, right] = -1
        patches_at_node[3, bottom] = -1

        return patches_at_node.T


class StructuredQuadGraphTopology:
    _layout = StructuredQuadLayoutCython

    def __init__(self, shape):
        self._shape = tuple(shape)

    @property
    def shape(self):
        return self._shape

    @property
    def number_of_node_rows(self):
        return self._shape[0]

    @property
    def number_of_node_columns(self):
        return self._shape[1]

    @cached_property
    @read_only_array
    def nodes(self):
        """A shaped array of node ids.

        Returns
        -------
        ndarray
            Node IDs in an array shaped as *number_of_node_rows* by
            *number_of_node_columns*.
        """
        return np.arange(self.shape[0] * self.shape[1]).reshape(self.shape)

    @cached_property
    @read_only_array
    def nodes_at_right_edge(self):
        return np.arange(self.shape[1] - 1, np.prod(self.shape), self.shape[1])

    @cached_property
    @read_only_array
    def nodes_at_top_edge(self):
        return np.arange(self.number_of_nodes - self.shape[1], np.prod(self.shape))

    @cached_property
    @read_only_array
    def nodes_at_left_edge(self):
        return np.arange(0, np.prod(self.shape), self.shape[1])

    @cached_property
    @read_only_array
    def nodes_at_bottom_edge(self):
        return np.arange(self.shape[1])

    def nodes_at_edge(self, edge):
        if edge not in ("right", "top", "left", "bottom"):
            raise ValueError("value for edge not understood")
        return getattr(self, f"nodes_at_{edge}_edge")

    @property
    def nodes_at_corners_of_grid(self):
        """Nodes at corners of grid.

        The nodes at at the corners of the grid. The nodes are returned
        counterclockwise starting with the upper-right.

        Return
        ------
        tuple of int
            Nodes at the four corners.

        Examples
        --------
        >>> from landlab.graph import UniformRectilinearGraph
        >>> graph = UniformRectilinearGraph((4, 5))
        >>> graph.nodes_at_corners_of_grid
        (19, 15, 0, 4)
        """
        return (
            self.number_of_nodes - 1,
            self.number_of_nodes - self.number_of_node_columns,
            0,
            self.number_of_node_columns - 1,
        )

    @cached_property
    @read_only_array
    def nodes_at_link(self):
        return self._layout.nodes_at_link(self.shape)

    @cached_property
    def horizontal_links(self):
        return self._layout.horizontal_links(self.shape)

    @cached_property
    def vertical_links(self):
        return self._layout.vertical_links(self.shape)

    @property
    def corner_nodes(self):
        n_rows, n_cols = self.shape
        return np.asarray(
            (n_rows * n_cols - 1, (n_rows - 1) * n_cols, 0, n_cols - 1), dtype=int
        )

    @cached_property
    def perimeter_nodes(self):
        return self._layout.perimeter_nodes(self.shape)

    @cached_property
    def links_at_node(self):
        return self._layout.links_at_node(self.shape)

    @cached_property
    def link_dirs_at_node(self):
        return self._layout.link_dirs_at_node(self.shape)

    @cached_property
    @read_only_array
    def patches_at_link(self):
        return self._layout.patches_at_link(self.shape)

    @cached_property
    @read_only_array
    def patches_at_node(self):
        return self._layout.patches_at_node(self.shape)


class StructuredQuadGraphExtras(StructuredQuadGraphTopology, Graph):
    def __init__(self, node_y_and_x, sort=False):
        StructuredQuadGraphTopology.__init__(self, node_y_and_x[0].shape)
        Graph.__init__(
            self,
            node_y_and_x,
            links=StructuredQuadLayoutCython.nodes_at_link(self.shape),
            patches=StructuredQuadLayoutCython.links_at_patch(self.shape),
            sort=sort,
        )

    @property
    def nodes_at_link(self):
        return self.ds["nodes_at_link"].values


class StructuredQuadGraph(StructuredQuadGraphExtras):
    def __init__(self, coords, shape=None, sort=False):
        node_y, node_x = (
            np.asarray(coords[0], dtype=float),
            np.asarray(coords[1], dtype=float),
        )

        if shape:
            node_y.shape = shape
            node_x.shape = shape
        else:
            node_x.shape = node_y.shape

        if node_y.shape != node_x.shape:
            raise ValueError("shape mismatch in node x and y coordinates")

        StructuredQuadGraphExtras.__init__(self, (node_y, node_x), sort=sort)

    @staticmethod
    def setup_node_y_and_x(yx_at_node, shape=None):
        node_y = np.asarray(yx_at_node[0], dtype=float)
        node_x = np.asarray(yx_at_node[1], dtype=float)

        if shape:
            node_y.shape = shape
            node_x.shape = shape
        else:
            node_x.shape = node_y.shape

        if node_y.shape != node_x.shape:
            raise ValueError("shape mismatch in node x and y coordinates")

        return (node_y, node_x)


class RectilinearGraph(StructuredQuadGraphExtras):

    """Graph of a rectlinear grid of nodes.

    Examples
    --------
    >>> from landlab.graph import RectilinearGraph
    >>> graph = RectilinearGraph(([0, 1, 2, 3], [1, 4, 8]))
    >>> graph.number_of_nodes
    12
    >>> graph.y_of_node # doctest: +NORMALIZE_WHITESPACE
    array([ 0., 0., 0.,
            1., 1., 1.,
            2., 2., 2.,
            3., 3., 3.])
    >>> graph.x_of_node # doctest: +NORMALIZE_WHITESPACE
    array([ 1., 4., 8.,
            1., 4., 8.,
            1., 4., 8.,
            1., 4., 8.])
    """

    def __init__(self, nodes, sort=False):
        rows = np.asarray(nodes[0], dtype=float)
        cols = np.asarray(nodes[1], dtype=float)
        node_y_and_x = np.meshgrid(rows, cols, indexing="ij")

        StructuredQuadGraphExtras.__init__(self, node_y_and_x, sort=sort)

    @staticmethod
    def setup_node_y_and_x(coords):
        rows = np.asarray(coords[0], dtype=float)
        cols = np.asarray(coords[1], dtype=float)

        return np.meshgrid(rows, cols, indexing="ij")


class UniformRectilinearGraph(StructuredQuadGraphExtras):

    """Graph of a structured grid of quadrilaterals.

    Examples
    --------
    >>> from landlab.graph import UniformRectilinearGraph
    >>> graph = UniformRectilinearGraph((4, 3), spacing=(1, 2), origin=(-1, 0))
    >>> graph.number_of_nodes
    12
    >>> graph.y_of_node # doctest: +NORMALIZE_WHITESPACE
    array([-1., -1., -1.,
            0.,  0.,  0.,
            1.,  1.,  1.,
            2.,  2.,  2.])
    >>> graph.x_of_node # doctest: +NORMALIZE_WHITESPACE
    array([ 0.,  2.,  4.,
            0.,  2.,  4.,
            0.,  2.,  4.,
            0.,  2.,  4.])
    >>> graph.links_at_node # doctest: +NORMALIZE_WHITESPACE
    array([[ 0,  2, -1, -1], [ 1,  3,  0, -1], [-1,  4,  1, -1],
           [ 5,  7, -1,  2], [ 6,  8,  5,  3], [-1,  9,  6,  4],
           [10, 12, -1,  7], [11, 13, 10,  8], [-1, 14, 11,  9],
           [15, -1, -1, 12], [16, -1, 15, 13], [-1, -1, 16, 14]])
    >>> graph.link_dirs_at_node # doctest: +NORMALIZE_WHITESPACE
    array([[-1, -1,  0,  0], [-1, -1,  1,  0], [ 0, -1,  1,  0],
           [-1, -1,  0,  1], [-1, -1,  1,  1], [ 0, -1,  1,  1],
           [-1, -1,  0,  1], [-1, -1,  1,  1], [ 0, -1,  1,  1],
           [-1,  0,  0,  1], [-1,  0,  1,  1], [ 0,  0,  1,  1]], dtype=int8)
    >>> graph.nodes_at_link # doctest: +NORMALIZE_WHITESPACE
    array([[ 0,  1], [ 1,  2], [ 0,  3], [ 1,  4], [ 2,  5],
           [ 3,  4], [ 4,  5], [ 3,  6], [ 4,  7], [ 5,  8],
           [ 6,  7], [ 7,  8], [ 6,  9], [ 7, 10], [ 8, 11],
           [ 9, 10], [10, 11]])
    >>> graph.links_at_patch # doctest: +NORMALIZE_WHITESPACE
    array([[ 3,  5,  2,  0], [ 4,  6,  3,  1],
           [ 8, 10,  7,  5], [ 9, 11,  8,  6],
           [13, 15, 12, 10], [14, 16, 13, 11]])
    >>> graph.nodes_at_patch # doctest: +NORMALIZE_WHITESPACE
    array([[ 4,  3,  0,  1], [ 5,  4,  1,  2],
           [ 7,  6,  3,  4], [ 8,  7,  4,  5],
           [10,  9,  6,  7], [11, 10,  7,  8]])
    """

    def __init__(self, shape, spacing=1.0, origin=0.0, sort=False):
        spacing = np.broadcast_to(spacing, 2)
        origin = np.broadcast_to(origin, 2)

        rows = np.arange(shape[0], dtype=float) * spacing[0] + origin[0]
        cols = np.arange(shape[1], dtype=float) * spacing[1] + origin[1]

        node_y_and_x = np.meshgrid(rows, cols, indexing="ij")

        StructuredQuadGraphExtras.__init__(self, node_y_and_x, sort=sort)

        self._spacing = tuple(spacing)
        self._origin = tuple(origin)

    @property
    def spacing(self):
        return self._spacing

    @property
    def origin(self):
        return self._origin

    @property
    def dx(self):
        return self._spacing[1]

    @property
    def dy(self):
        return self._spacing[0]



#!/usr/bin/env python3

import numpy as np
import xarray as xr


class DataRecord:
    """Data structure to store variables in time and/or space dimensions.

    This class uses a xarray Dataset to store variables. This datastructure is
    located at the property ``dataset``. The DataRecord expands xarray Dataset
    with additional attributes and functions, including the ability to
    aggregate values on Landlab grid elements.

    DataRecord uses the concept of an "item", a physical thing that is located
    on the grid and has some properties, stored as variables. DataRecord tracks
    variables through time, variables associated with a Landlab grid across
    items, or both.

    Thus data variables can vary along one or both of the following dimensions:

        - time (model time)
        - item_id: variables can characterize a set of items (each identified
          by an individual id) that reside on the grid.

    If an item or set of items is defined, each item must be defined by the
    grid element and the element id at which it resides, e.g.:

        grid_element = 'node'
        element_id = 9.

    When items are defined, each item is given a unique id and the underlying
    Dataset uses a dimension "item_id". **Items are assigned ids beginning with
    0 followed by consecutively increasing integers.**

    Examples:

        - the variable 'mean_elevation' characterizes the grid and varies with
          time,
        - the variable 'clast__rock_type' characterizes a set of items (clasts)
          and varies with item_id,
        - the variable 'clast__size' can vary with both time and item_id

    In the above case, `grid_element` and `element_id` are default data
    variables (in addition to any user-specified variables).

    For each item, `element_id` must be less than the number of this item's
    grid_element that exist on the grid or be one of the dummy element values.
    For example, if the grid has 100 links, and no dummy link values are
    indicated, then, no item can live at link 100 or link -3 because only links
    0 to 99 exist in this example.

    Anything that the DataRecord keeps track of is considered a "record",
    whether it uses one or both of the two standard dimensions (**time** and
    **item_id**).

    DataRecord provides two method to assist with adding new records. The
    method ``add_item`` should be used when no new variables are being added.
    The method ``add_record`` should be used when new variables are being
    added or when a variable is only tracked over the **time** dimension.
    """

    _name = "DataRecord"

    def __init__(
        self,
        grid,
        dummy_elements=None,
        time=None,
        items=None,
        data_vars=None,
        attrs=None,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
        dummy_elements : dict
            Dictionary indicating valid values for dummy grid elements. For
            example, if you need an "exit" off of a grid with  100 links, you
            could indicate `dummy_elements = {"link": [9999]}`
            to set a link id of 9999 as a dummy link. Multiple dummy elements
            are possible and we recommend using values larger than the number
            of grid elements for the dummy values.
        time : list or 1-D array of float or int (optional)
            The initial time(s) to add to the record. A time dimension is not
            created if the value is 'None' (default).
        items : dict (optional)
            Generic items that live on grid elements. No item is created if the
            value is 'None' (default). Otherwise, dictionary describes the
            position of generic items on the grid. The structure is:

            .. code-block:: python

                {'grid_element' : [grid_element],
                 'element_id' : [element_id]}

            where:

                - [grid_element] is a str or number-of-items-long array
                  containing strings of the grid element(s) on which the items
                  live. Valid locations depend on the grid type. If provided as a
                  string it is assumed that all items live on the same type of
                  grid element.
                - [element_id] is an array of integers identifying the grid
                  element ID on which each item resides.

            An example argument would be:

            .. code-block:: python

                {'grid_element' : numpy.array(['node'], ['node'], ['link']),
                 'element_id' :   numpy.array([1],      [5],      [1]     )}

        data_vars : dict (optional)
            Dictionary of the data variables to be recorded. The structure is:

            .. code-block:: python

                {'variable_name_1' : (['dimensions'], variable_data_1),
                 'variable_name_2' : (['dimensions'], variable_data_2)}

            where:

                - 'variable_name...' is a string of the variable name (label)
                - ['dimensions'] is the dimension(s) over which the variable
                  exists: can be ['time'], ['item_id'] or ['item_id', 'time'].
                - variable_data is an array containing the data, its size must
                  match that of the variable dimension(s).
        attrs : dict (optional)
            Dictionary of global attributes on the DataRecord (metadata).
            Example: {'time_units' : 'y'}

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with time as the only dimension:

        >>> dr1 = DataRecord(grid, time=[0.],
        ...                  data_vars={'mean_elevation' : (['time'],
        ...                                                 np.array([100]))},
        ...                  attrs={'time_units' : 'y'})

        DataRecord builds off of xarray Dataset, a multi-dimensional, in
        memory, array  database. Dataset implements the mapping interface with
        keys given by variable names and values given by DataArray objects for
        each variable name.

        A DataRecord can have dimensions 'time' and/or 'item_id'.

        The xarray Dataset is stored in the public attribute ``dataset``.

        Coordinates are one dimensional arrays used for label-based indexing.
        DataRecord inherits all the methods and attributes from
        ``xarray.Dataset``.

        >>> dr1.dataset.to_dataframe()
              mean_elevation
        time
        0.0              100
        >>> dr1.dataset.time.values
        array([ 0.])
        >>> dr1.variable_names
        ['mean_elevation']
        >>> dr1.dataset['mean_elevation'].values
        array([100])
        >>> list(dr1.dataset.attrs.items())
        [('time_units', 'y')]

        >>> list(dr1.dataset.attrs.items())
        [('time_units', 'y')]

        Example of a DataRecord with item_id as the only dimension:

        >>> my_items2 = {
        ...     'grid_element': np.array(('node', 'link'), dtype=str),
        ...     'element_id': np.array([1, 3]),
        ... }
        >>> dr2 = DataRecord(grid, items=my_items2)

        Note that both arrays (grid_element and element_id) have 1 dimension
        as they only vary along the dimension 'item_id'.

        >>> dr2.dataset.to_dataframe()[['grid_element', 'element_id']]
                grid_element  element_id
        item_id
        0               node           1
        1               link           3

        Example of a DataRecord with dimensions time and item_id:

        >>> my_items3 = {'grid_element':np.array([['node'], ['link']]),
        ...              'element_id': np.array([[1], [3]])}
        >>> dr3 = DataRecord(grid, time=[0.], items=my_items3)

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.

        >>> dr3.dataset.to_dataframe()[['grid_element', 'element_id']]
                     grid_element  element_id
        item_id time
        0       0.0          node           1
        1       0.0          link           3

        """

        # save a reference to the grid
        self._grid = grid

        # depending on the grid type, permitted locations for items vary
        self._permitted_locations = self._grid.groups

        # save dummy elements reference
        # check dummies and reformat into {"node": [0, 1, 2]}
        self._dummy_elements = dummy_elements or {}
        for at in self._permitted_locations:
            for item in self._dummy_elements.get(at, []):
                if (item < self._grid[at].size) and (item >= 0):
                    raise ValueError(f"Dummy id {at} {item} invalid")

        # set initial time coordinates, if any
        if isinstance(time, (list, np.ndarray)):
            self._times = np.array(time)
            self._number_of_times = len(self._times)
        elif time is not None:
            raise TypeError("time must be a list or numpy array")

        # set initial items, if any
        if items is not None:
            try:
                items.keys()
            except AttributeError as exc:
                # items is not a dict
                raise TypeError(
                    "You must provide an `items` dictionary "
                    "(see documentation for required format)"
                ) from exc
            try:
                _grid_elements, _element_ids = (
                    items["grid_element"],
                    items["element_id"],
                )
            except KeyError as exc:
                # grid_element and/or element_id not provided
                raise TypeError(
                    "You must provide an `items` dictionary,"
                    "(see documentation for required format)"
                ) from exc

            self._number_of_items = len(_element_ids)
            if len(_grid_elements) != self._number_of_items:
                if isinstance(_grid_elements, str):
                    pass
                else:
                    raise ValueError(
                        "The number of grid_element passed "
                        "to DataRecord must be 1 or equal "
                        "to the number of element_id."
                    )

            # check that grid_element and element_id exist on the grid and
            # have valid format:
            _grid_elements, _element_ids = self._check_grid_element_and_id(
                _grid_elements, _element_ids
            )

            # check that element IDs do not exceed number of elements
            # on the grid:
            self._check_element_id_values(_grid_elements, _element_ids)

            # create coordinates for the dimension 'item_id':
            self._item_ids = np.array(range(self._number_of_items), dtype=int)

            # create initial dictionaries of variables:
            if time is not None:
                data_vars_dict = {
                    "grid_element": (["item_id", "time"], _grid_elements),
                    "element_id": (
                        ["item_id", "time"],
                        _element_ids,
                        {"dtype": int},
                    ),
                }
                coords = {"time": self._times, "item_id": self._item_ids}
            else:
                # no time
                data_vars_dict = {
                    "grid_element": (["item_id"], _grid_elements),
                    "element_id": (["item_id"], _element_ids, {"dtype": int}),
                }
                coords = {"item_id": self._item_ids}

        else:
            # no items, initial dictionary of variables is empty:
            data_vars_dict = {}
            if time is not None:
                coords = {"time": self._times}
            else:  # no item and no time = no dimension
                coords = {}

        # set variables, if any
        if data_vars is not None:
            try:
                # check format (dict)
                data_vars.keys()
            except AttributeError as exc:
                raise TypeError(
                    "Data variables (data_vars) passed to "
                    "DataRecord must be a dictionary (see "
                    "documentation for valid structure)"
                ) from exc
            for key in data_vars.keys():
                # check dict structure and dims:
                if data_vars[key][0] not in (
                    ["time"],
                    ["item_id"],
                    ["time", "item_id"],
                    ["item_id", "time"],
                ):
                    raise ValueError(
                        "Data variable dimensions must be " "time and/or item_id"
                    )

            # create complete dictionary of variables
            # (= initial data_vars_dict + additional user-defined data_vars):
            data_vars_dict.update(data_vars)

        # set attributes, if any
        attrs = attrs or {}
        if not isinstance(attrs, dict):
            raise TypeError(
                "Attributes (attrs) passed to DataRecord" "must be a dictionary"
            )

        # create an xarray Dataset:
        self._dataset = xr.Dataset(data_vars=data_vars_dict, coords=coords, attrs=attrs)

    def _check_grid_element_and_id(self, grid_element, element_id):
        """Check the location and size of grid_element and element_id."""
        if isinstance(grid_element, str):
            # create list of grid_element for all items
            ge_name = grid_element
            if hasattr(self, "_number_of_times"):
                # if time
                grid_element = np.array(
                    np.empty(
                        (self._number_of_items, self._number_of_times), dtype=object
                    )
                )

                if element_id.shape != grid_element.shape:
                    element_id = np.broadcast_to(element_id, grid_element.shape)

            else:
                # no time
                grid_element = np.array(
                    np.empty((self._number_of_items,), dtype=object)
                )
            grid_element.fill(ge_name)

        # verify all grid elements are valid.
        for loc in grid_element.flatten():
            if loc not in self._permitted_locations:
                raise ValueError(
                    "One or more of the grid elements"
                    " provided is/are not permitted location"
                    " for this grid type"
                )

        return grid_element, element_id

    def _check_element_id_values(self, grid_element, element_id):
        """Check that element_id values are valid."""
        for at in self._permitted_locations:
            max_size = self._grid[at].size

            # this needs to work with 2d arrays (rows, col = np.where (so grid
            # element always needs to be at least 2d.))
            ind = np.nonzero(grid_element == at)
            selected_elements = element_id[ind]

            if selected_elements.size > 0:
                dummy_values = self._dummy_elements.get(at, [])
                index_values = np.arange(0, max_size)
                valid_values = np.concatenate((dummy_values, index_values))

                valid_elements = np.isin(selected_elements, valid_values)

                if not np.all(valid_elements):
                    raise ValueError("Invalid element_ids provided.")

        if not np.issubdtype(element_id.dtype, np.integer):
            raise ValueError(
                "You have passed a non-integer element_id to "
                "DataRecord, this is not permitted"
            )

    def add_record(self, time=None, item_id=None, new_item_loc=None, new_record=None):
        """Add a new record to the DataRecord.

        Unlike add_item, this method can support adding records that include
        new variables to the DataRecord. It can also support adding records
        that do not include time.

        Parameters
        ----------
        time : list or 1-D array of float or int
            Time step at which the record is to be added.
        item_id : list or 1-D array of int (optional)
            ID of the item to which the new record relates.
        new_item_loc: dict (optional)
            Dictionary of the new item location. If the new record is a change
            in the item location (grid_element and/or element_id), this field
            must be provided as:

            .. code-block:: python

                {'grid_element' : [grid_element],
                 'element_id' : [element_id]}

            Both must be provided even if only one is being changed.

        new_record : dict
            Dictionary containing the new record. Structure should be:
            {'variable_name_1' : (['dimensions'], variable_data_1)}
            with:

                - 'variable_name_1' : name of the (potentially new) variable
                - ['dimensions'] : dimension(s) along which the new record
                  varies; can be ['time'], ['item_id] or ['item_id', 'time']
                - variable_data_1 : new data array, size must match the
                  variable dimension(s)

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with dimensions time and item_id:

        >>> my_items3 = {
        ...     "grid_element": np.array([["node"], ["link"]]),
        ...     "element_id": np.array([[1],[3]]),
        ... }

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.

        >>> dr3 = DataRecord(grid, time=[0.], items=my_items3)

        Records relating to pre-existing items can be added to the DataRecord
        using the method 'add_record':

        >>> dr3.add_record(
        ...     time=[2.0],
        ...     item_id=[0],
        ...     new_item_loc={
        ...         'grid_element' : np.array([['node']]),
        ...         'element_id' : np.array([[6]])
        ...     },
        ...     new_record={
        ...         'item_size':(['item_id', 'time'], np.array([[0.2]]))
        ...     }
        ... )
        >>> dr3.dataset['element_id'].values
        array([[  1.,   6.],
               [  3.,  nan]])
        >>> dr3.get_data([2.0],[0],'item_size')
        array([ 0.2])

        The 'add_record' method can also be used to add a non item-related
        record:

        >>> dr3.add_record(time=[50.0], new_record={'mean_elev': (['time'], [110])})
        >>> dr3.dataset['mean_elev'].to_dataframe()
              mean_elev
        time
        0.0         NaN
        2.0         NaN
        50.0      110.0
        """
        if time is not None:
            try:
                # check that time is a dim of the DataRecord
                self._dataset["time"]
            except KeyError as exc:
                raise KeyError("This DataRecord does not record time") from exc

            if not isinstance(time, (list, np.ndarray)):
                # check input type
                raise TypeError(
                    "You have passed a time that is"
                    " not permitted, must be list or array"
                )
            else:
                if item_id is not None:
                    try:
                        # check that DataRecord holds items
                        self._dataset["item_id"]
                    except KeyError as exc:
                        raise KeyError("This DataRecord does not hold items") from exc
                    try:
                        # check that item_id is list or array
                        len(item_id)
                    except TypeError as exc:
                        raise TypeError("item_id must be a list or a 1D array") from exc
                    if not all(i in self._dataset["item_id"].values for i in item_id):
                        # check that item_id already exist
                        raise ValueError(
                            "One or more of the value(s) you "
                            "passed as item_id is/are not "
                            "currently in the DataRecord. Change"
                            " the input values create a new item"
                            "using the method add_item"
                        )
                    coords_to_add = {"time": np.array(time), "item_id": item_id}

                    # if item location is changed by this new record, check
                    # that both grid_element and element_id are provided:
                    if new_item_loc is not None:
                        try:
                            new_grid_element = new_item_loc["grid_element"]
                            new_element_id = new_item_loc["element_id"]
                        except KeyError as exc:
                            raise KeyError(
                                "You must provide a "
                                "new_item_loc dictionary with both "
                                "grid_element and element_id"
                            ) from exc
                        # check that grid_element and element_id exist
                        # on the grid and have valid format:
                        (
                            new_grid_element,
                            new_element_id,
                        ) = self._check_grid_element_and_id(
                            new_grid_element, new_element_id
                        )

                        # check that element IDs do not exceed number
                        # of elements on this grid:
                        self._check_element_id_values(new_grid_element, new_element_id)

                        _new_data_vars = {
                            "grid_element": (["item_id", "time"], new_grid_element),
                            "element_id": (["item_id", "time"], new_element_id),
                        }
                    else:
                        # new_item_loc is `None`
                        _new_data_vars = {}
                else:
                    # no item
                    coords_to_add = {"time": np.array(time)}
                    _new_data_vars = {}

        else:
            # no time
            if item_id is not None:
                if not all(i in self._dataset["item_id"].values for i in item_id):
                    # check that item_id already exist
                    raise ValueError(
                        "One or more of the value(s) you "
                        "passed as item_id is/are not "
                        "currently in the DataRecord. Change"
                        " the input values create a new item"
                        "using the method add_item"
                    )

                coords_to_add = {"item_id": np.array(item_id)}
                _new_data_vars = {}

                # no time so if item location needs to be changed,
                # user should use set_data
                if new_item_loc is not None:
                    raise ValueError(
                        "Use the method set_data to change the "
                        "location of an item in this DataRecord"
                    )
            else:
                # no item
                _new_data_vars = {}
                coords_to_add = {}

        if new_record is not None:
            # add new_record to dict of variables to add
            _new_data_vars.update(new_record)

        # create dataset of new record
        ds_to_add = xr.Dataset(data_vars=_new_data_vars, coords=coords_to_add)

        # merge new record and original dataset
        self._dataset = xr.merge((self._dataset, ds_to_add), compat="no_conflicts")

    def add_item(self, time=None, new_item=None, new_item_spec=None):
        """Add new item(s) to the current DataRecord.

        Parameters
        ----------
        time : list or 1-D array of float or int
            Time step at which the items are to be added.
        new_item : dict
            Structure is:

            .. code-block:: python

                {'grid_element' : [grid_element],
                 'element_id' : [element_id]}

            where:

                - [grid_element] is str or number-of-items long array
                  containing strings of the grid element(s) on which the items
                  live. Valid locations depend on the grid type. If provided as
                  a string it is assumed that all items live on the same type of
                  grid element.
                - [element_id] is an array of integers identifying the grid
                  element ID on which each item resides.

            An example argument would be:

            .. code-block:: python

                {'grid_element' : numpy.array([['node'], ['node'], ['link']]),
                 'element_id' :   numpy.array([[1],      [5],      [1]     ])}

        new_item_spec : dict (optional)
            Dictionary containing any data variables (other than
            'grid_element' and 'element_id') relating to the new item(s) to be
            added. Structure is:

            .. code-block:: python

                {'variable_name_1' : (['dimensions'], variable_data_1)}

            where:

                - 'variable_name_1' : name of the (potentially new) variable
                - ['dimensions'] : dimension(s) along which the new record
                  varies; can be ['time'], ['item_id] or ['item_id', 'time']
                - variable_data_1 : new data array, size must match the
                  variable dimension(s)

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with dimensions time and item_id:

        >>> my_items3 = {'grid_element':np.array([['node'], ['link']]),
        ...              'element_id': np.array([[1],[3]])}

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.

        >>> dr3 = DataRecord(grid,
        ...                  time=[0.],
        ...                  items=my_items3)

        Items can be added to a DataRecord that already holds similar items,
        using the method 'add_item':

        >>> dr3.add_item(time=[1.0],
        ...              new_item={'grid_element' : np.array(
        ...                                              [['node'], ['node']]),
        ...                        'element_id' : np.array([[4],[4]])},
        ...              new_item_spec={'size': (
        ...                              ['item_id', 'time'], [[10],[5]])})

        Two items have been added at a new timestep 1.0:

        >>> dr3.number_of_items
        4
        >>> dr3.time_coordinates
        [0.0, 1.0]

        If a data variable is also added with the new items ('size' in this
        example), the values for this variable are filled with 'nan' for the
        pre-existing items:

        >>> dr3.dataset['size'][:,1].values
        array([ nan,  nan,  10.,   5.])

        The previous line calls the values of the variable 'size', for all
        items, at time=1; the first two items don't have a value for the
        variable 'size'.
        """
        if time is None and "time" in self._dataset["grid_element"].coords:
            raise ValueError(
                "The items previously defined in this DataRecord"
                ' have dimensions "time" and "item_id", '
                'you must provide a "time" for the new item(s)'
            )

        if not isinstance(new_item, dict):
            raise TypeError(
                "You must provide an new_item dictionary "
                "(see documentation for required format)"
            )

        try:
            # check that dict contains correct entries
            _grid_elements, _element_ids = (
                new_item["grid_element"],
                new_item["element_id"],
            )

        except KeyError as exc:
            raise KeyError(
                "You must provide a new_item dictionary "
                "(see documentation for required format)"
            ) from exc

        number_of_new_items = len(new_item["element_id"])
        # first id of new item = last item in existing datarecord+1
        new_first_item_id = self._dataset["item_id"][-1].values + 1
        new_item_ids = np.array(
            range(new_first_item_id, new_first_item_id + number_of_new_items)
        )

        if time is not None:
            try:
                self._dataset["time"]
            except KeyError as exc:
                raise KeyError("This DataRecord does not record time") from exc
            if not isinstance(time, (list, np.ndarray)):
                raise TypeError(
                    "You have passed a time that is not "
                    "permitted, must be list or a 1-D array"
                )
            else:
                coords_to_add = {
                    "time": np.array(time),
                    "item_id": np.array(new_item_ids),
                }
                # check that grid_element and element_id exist
                # on the grid and have valid format
                _grid_elements, _element_ids = self._check_grid_element_and_id(
                    _grid_elements, _element_ids
                )

                # check that element IDs do not exceed number
                # of elements on this grid
                self._check_element_id_values(_grid_elements, _element_ids)

                data_vars_dict = {
                    "grid_element": (["item_id", "time"], _grid_elements),
                    "element_id": (["item_id", "time"], _element_ids),
                }

        else:
            # no time
            coords_to_add = {"item_id": np.array(new_item_ids)}
            # check that grid_element and element_id exist on
            # the grid and have valid format:
            _grid_elements, _element_ids = self._check_grid_element_and_id(
                _grid_elements, _element_ids
            )
            # check that element IDs do not exceed number of
            # elements on this grid
            self._check_element_id_values(_grid_elements, _element_ids)

            data_vars_dict = {
                "grid_element": (["item_id"], _grid_elements),
                "element_id": (["item_id"], _element_ids),
            }

        # other variables:
        if new_item_spec is not None:
            data_vars_dict.update(new_item_spec)

        # Dataset of new record:
        ds_to_add = xr.Dataset(data_vars=data_vars_dict, coords=coords_to_add)

        # Merge new record and original dataset:
        self._dataset = xr.merge((self._dataset, ds_to_add), compat="no_conflicts")

    def get_data(self, time=None, item_id=None, data_variable=None):
        """Get the value of a variable at a model time and/or for an item.

        Parameters
        ----------
        time : list or 1-D array of float or int (optional)
            The time coordinate of the record to get.
        item_id : list or 1-D array of int (optional)
            The item id of the record to get.
        data_variable : string
            The label of the variable to get.

        Returns
        -------
        object
            The value of *variable* at *time* and/or for *item_id*. The type of
            the returned object is dependent on the type of the variable value.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with dimensions time and item_id:

        >>> my_items4 = {'grid_element' : 'node',
        ...              'element_id': np.array([[1],[3],[3],[7]])}

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.

        >>> my_data4 = {'item_size': (['item_id', 'time'], np.array(
        ...                                 [[0.3],[0.4],[0.8],[0.4]]))}
        >>> dr4=DataRecord(grid,
        ...                time=[50.],
        ...                items=my_items4,
        ...                data_vars=my_data4)
        >>> dr4.get_data([50.],[2],'element_id')
        array([3])
        >>> dr4.get_data(time=[50.],data_variable='item_size')
        array([  0.3,  0.4,  0.8,  0.4])
        >>> dr4.get_data(item_id=[1,2], data_variable='grid_element')
        array([['node'],
               ['node']], dtype=object)
        """
        try:
            self._dataset[data_variable]
        except KeyError as exc:
            raise KeyError(
                f"the variable {data_variable!r} is not in the DataRecord"
            ) from exc
        if time is None:
            if item_id is None:
                return self._dataset[data_variable].values
            else:
                try:
                    self._dataset["item_id"]
                except KeyError as exc:
                    raise KeyError("This DataRecord does not hold items") from exc
                try:
                    len(item_id)
                except TypeError as exc:
                    raise TypeError("item_id must be a list or a 1-D array") from exc
                try:
                    self._dataset["item_id"].values[item_id]
                except IndexError as exc:
                    raise IndexError(
                        "The item_id you passed does not exist " "in this DataRecord"
                    ) from exc

                return self._dataset.isel(item_id=item_id)[data_variable].values

        else:  # time is not None
            try:
                self._dataset["time"]
            except KeyError as exc:
                raise KeyError("This DataRecord does not record time") from exc
            try:
                len(time)
            except TypeError as exc:
                raise TypeError("time must be a list or a 1-D array") from exc
            try:
                time_index = int(self.time_coordinates.index(time[0]))
            except ValueError as exc:
                raise IndexError(
                    "The time you passed is not currently"
                    " in the DataRecord, you must change the value"
                    " you pass or first create the new time "
                    " coordinate using the add_record method"
                ) from exc
            if item_id is None:
                return self._dataset.isel(time=time_index)[data_variable].values
            else:
                try:
                    self._dataset["item_id"]
                except KeyError as exc:
                    raise KeyError("This DataRecord does not hold items") from exc
                try:
                    len(item_id)
                except TypeError as exc:
                    raise TypeError("item_id must be a list or a 1-D array") from exc
                try:
                    self._dataset["item_id"].values[item_id]
                except IndexError as exc:
                    raise IndexError(
                        "The item_id you passed does not exist " "in this DataRecord"
                    ) from exc
                return self._dataset.isel(time=time_index, item_id=item_id)[
                    data_variable
                ].values

    def set_data(self, time=None, item_id=None, data_variable=None, new_value=np.nan):
        """Set a variable value at a model time and/or an item to a new value.

        The value of only one variable can be changed at a time using this
        method.

        Parameters
        ----------
        time : list or 1-D array of float or int
            The time coordinate of the record to set.
        item_id : list or 1-D array of int
            The item id of the record to set.
        data_variable : string
            The label of the variable to set.
        new_value : list or 1-D array
            The new value to give to the variable data.

        Returns
        -------
        DataRecord with updated data.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with dimensions time and item_id:

        >>> my_items4 = {'grid_element' : 'node',
        ...              'element_id': np.array([[1],[3],[3],[7]])}

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.

        >>> my_data4 = {'item_size': (['item_id', 'time'], np.array(
        ...                                 [[0.3],[0.4],[0.8],[0.4]]))}
        >>> dr4 = DataRecord(grid,
        ...                  time=[50.],
        ...                  items=my_items4,
        ...                  data_vars=my_data4)
        >>> dr4.dataset['item_size'].values
        array([[ 0.3],
               [ 0.4],
               [ 0.8],
               [ 0.4]])
        >>> dr4.set_data([50.],[2],'item_size', [0.5])
        >>> dr4.dataset['item_size'].values
        array([[ 0.3],
               [ 0.4],
               [ 0.5],
               [ 0.4]])
        """
        if data_variable not in self.variable_names:
            raise KeyError(
                "the variable '{}' is not in the " "DataRecord".format(data_variable)
            )

        # If record to be changed is 'grid_element' or 'element_id',
        # check that provided grid_element is valid and that new
        # grid_element+element_id combination exist on the grid and
        # have valid format:
        if data_variable in ("grid_element", "element_id"):
            if data_variable == "grid_element":
                assoc_grid_element = new_value
                assoc_element_id = self.get_data(time, item_id, "element_id")[0]
            if data_variable == "element_id":
                if not isinstance(new_value, int):
                    raise ValueError(
                        "You have passed a non-integer "
                        "element_id to DataRecord, this is not "
                        "permitted"
                    )
                if new_value < 0:
                    raise ValueError(
                        "You have passed an element id below "
                        "zero. This is not permitted"
                    )
                assoc_element_id = new_value
                assoc_grid_element = self.get_data(time, item_id, "grid_element")[0]
            self._check_grid_element_and_id(assoc_grid_element, assoc_element_id)
            if assoc_element_id >= self._grid[assoc_grid_element].size:
                raise ValueError(
                    "The location "
                    + assoc_grid_element
                    + " "
                    + str(assoc_element_id)
                    + " does not exist on this grid"
                )

        if time is None:
            self._dataset[data_variable].values[item_id] = new_value
        else:
            try:
                len(time)
            except TypeError as exc:
                raise TypeError("time must be a list or a 1-d array") from exc
            try:
                # check that time coordinate already exists
                time_index = np.where(self._dataset.time.values == time)[0][0]
            except IndexError as exc:
                raise IndexError(
                    "The time you passed is not currently"
                    " in the DataRecord, you must change the value"
                    " you pass or first create the new time "
                    " coordinate using the add_record method"
                ) from exc

            if item_id is None:
                self._dataset[data_variable].values[time_index] = new_value
            else:
                try:
                    len(item_id)
                except TypeError as exc:
                    raise TypeError("item_id must be a list or a 1-d array") from exc
                try:
                    self._dataset["item_id"]
                    self._dataset[data_variable].values[item_id, time_index] = new_value
                except KeyError as exc:
                    raise KeyError("This DataRecord does not hold items") from exc

    def calc_aggregate_value(
        self,
        func,
        data_variable,
        at="node",
        filter_array=None,
        fill_value=np.nan,
        args=(),
        **kwargs,
    ):
        """Apply a function to a variable aggregated at grid elements.

        Parameters
        ----------
        func : function
            Function to apply to be aggregated.
        data_variable : str
            Name of variable on which to apply the function.
        at : str, optional
            Name of grid element at which to apply the function.
            Default is "node".
        filter_array: boolean array with dimensions matching that of the
            DataRecord (optional)
            Array to filter the DataRecord before aggregation.
        fill_value: float
            Fill value for array. Default is np.nan.
        args : tuple (optional)
            Additional positional arguments to pass to the function.
        **kwargs : key value pairs (optional)
            Additional keyword arguments to pass to func.

        Returns
        -------
        out : ndarray
            Array of size number-of-grid_elements (grid_elements is the group
            passed as 'at' argument).

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.data_record import DataRecord
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((3,3))
        >>> element_id = [0, 0, 0, 0, 1, 2, 3, 4, 5, 9999]
        >>> volumes = [4, 5, 1, 2, 3, 4, 5, 6, 7, 1234]
        >>> ages = [10, 11, 12, 13, 14, 15, 16, 8, 10, 3456]
        >>> grid_element = 'node'
        >>> data = {'ages': ages,
        ...         'volumes': volumes}
        >>> dr = DataRecord(grid,
        ...                 dummy_elements={"node": [9999]},
        ...                 items={'grid_element' : 'node',
        ...                           'element_id' : np.array(element_id)},
        ...                 data_vars={'ages' : (['item_id'], np.array(ages)),
        ...                             'volumes' : (
        ...                                 ['item_id'], np.array(volumes))})
        >>> s = dr.calc_aggregate_value(func=xr.Dataset.sum, data_variable="ages")
        >>> s
        array([ 46.,  14.,  15.,  16.,   8.,  10.,  nan,  nan,  nan])
        >>> len(s) == grid.number_of_nodes
        True

        If you want to first filter the DataRecord and then aggregate, first
        create a filter array with dimensions matching that of the DataRecord
        and has `True` for entries that should be retained and False for
        entries that should be ignored.

        For example, if we wanted to aggregate volume for items with an age
        greater than 10 we would to the following:

        >>> f = dr.dataset['ages'] > 10.
        >>> v_f = dr.calc_aggregate_value(
        ...     func=xr.Dataset.sum, data_variable='volumes', filter_array=f
        ... )
        >>> v_f
        array([  8.,   3.,   4.,   5.,  nan,  nan,  nan,  nan,  nan])

        If we wanted the value for elements with no volume to be zero instead
        of np.nan we could use the keyword argument ``fill_value``.

        >>> f = dr.dataset['ages'] > 10.
        >>> v_f = dr.calc_aggregate_value(
        ...     func=xr.Dataset.sum,
        ...     data_variable="volumes",
        ...     filter_array=f,
        ...     fill_value=0.,
        ... )
        >>> v_f
        array([  8.,   3.,   4.,   5.,  0.,  0.,  0.,  0.,  0.])

        An array of ``fill_value`` is returned when ``filter_array`` is all
        ``False`` (np.nan is the default value).

        >>> f = dr.dataset['ages'] > 4000.
        >>> v_f = dr.calc_aggregate_value(
        ...     func=xr.Dataset.sum, data_variable="volumes", filter_array=f
        ... )
        >>> v_f
        array([  nan,   nan,   nan,   nan,  nan,  nan,  nan,  nan,  nan])

        Other values can be specified for ``fill_value``.

        >>> f = dr.dataset['ages'] > 4000.
        >>> v_f = dr.calc_aggregate_value(
        ...     func=xr.Dataset.sum,
        ...     data_variable="volumes",
        ...     filter_array=f,
        ...     fill_value=0.,
        ... )
        >>> v_f
        array([  0.,   0.,   0.,   0.,  0.,  0.,  0.,  0.,  0.])
        """
        filter_at = self._dataset["grid_element"] == at

        filter_valid_element = (self._dataset["element_id"] >= 0) * (
            self._dataset["element_id"] < self._grid[at].size
        )

        if filter_array is None:
            my_filter = filter_at * filter_valid_element
        else:
            my_filter = filter_at * filter_valid_element * filter_array

        if np.any(my_filter):
            # Filter DataRecord with my_filter and groupby element_id:
            filtered = self._dataset.where(my_filter).groupby("element_id")

            # Calculate values
            vals = filtered.map(func, *args, **kwargs)  # .reduce

            # Create a nan array that we will fill with the results of the sum
            # this should be the size of the number of elements, even if there are
            # no items living at some grid elements.
            out = fill_value * np.ones(self._grid[at].size)

            # put the values of the specified variable into the correct location
            # of the out array.
            out[vals.element_id.values.astype(int)] = vals[data_variable]

            return out
        else:
            return np.repeat(fill_value, self._grid[at].size)

    def ffill_grid_element_and_id(self):
        """Fill NaN values of the fields 'grid_element' and 'element_id'.

        Fields are filled by propagating values forward in time.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with dimensions time and item_id:

        >>> my_items3 = {'grid_element': np.array([['node'], ['link']]),
        ...              'element_id': np.array([[1],[3]])}

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.

        >>> dr3 = DataRecord(grid,
        ...                  time=[0.],
        ...                  items=my_items3)

        Records relating to pre-existing items can be added to the DataRecord
        using the method 'add_record':

        >>> dr3.add_record(time=[2.0, 3.0],
        ...                new_record={'mean_elevation':(
        ...                           ['time'], np.array([200., 250.]))})

        Adding this data record created two new time coordinates. The
        grid_element and element_id of the items has been filled with 'nan'
        for these time coordinates.

        >>> dr3.dataset['grid_element'].values
        array([['node', nan, nan],
               ['link', nan, nan]], dtype=object)
        >>> dr3.dataset['element_id'].values
        array([[  1.,  nan,  nan],
               [  3.,  nan,  nan]])

        To fill these values with the last valid value, use the method
        ffill_grid_element_and_id:

        >>> dr3.ffill_grid_element_and_id()
        >>> dr3.dataset['grid_element'].values
        array([['node', 'node', 'node'],
               ['link', 'link', 'link']], dtype=object)
        >>> dr3.dataset['element_id'].values
        array([[ 1.,  1.,  1.],
               [ 3.,  3.,  3.]])

        In some applications, there may be no prior valid value. Under these
        circumstances, those values will stay as NaN. That is, this only
        forward fills, and does not backfill.

        >>> my_items3 = {'grid_element':np.array([['node'], ['link']]),
        ...              'element_id': np.array([[1],[3]])}
        >>> dr3 = DataRecord(grid,
        ...                  time=[0.],
        ...                  items=my_items3)
        >>> dr3.dataset['element_id'].values
        array([[1], [3]])
        >>> dr3.dataset['grid_element'].values
        array([['node'],
               ['link']],
              dtype='<U4')

        Next add some new items at a new time.

        >>> dr3.add_item(time=[1.0],
        ...              new_item={'grid_element' : np.array(
        ...                                              [['node'], ['node']]),
        ...                        'element_id' : np.array([[4],[4]])},
        ...              new_item_spec={'size': (
        ...                              ['item_id', 'time'], [[10],[5]])})

        Two items have been added at a new timestep 1.0:

        >>> dr3.number_of_items
        4
        >>> dr3.time_coordinates
        [0.0, 1.0]
        >>> dr3.dataset['element_id'].values
        array([[  1.,  nan],
               [  3.,  nan],
               [ nan,   4.],
               [ nan,   4.]])
        >>> dr3.dataset['grid_element'].values
        array([['node', nan],
               ['link', nan],
               [nan, 'node'],
               [nan, 'node']], dtype=object)

        We expect that the NaN's to the left of the 4.s will stay NaN. And they
        do.

        >>> dr3.ffill_grid_element_and_id()
        >>> dr3.dataset['element_id'].values
        array([[  1.,   1.],
               [  3.,   3.],
               [ nan,   4.],
               [ nan,   4.]])
        >>> dr3.dataset['grid_element'].values
        array([['node', 'node'],
               ['link', 'link'],
               [nan, 'node'],
               [nan, 'node']], dtype=object)

        Finally, if we add a new time, we see that we need to fill in the
        full time column.

        >>> dr3.add_record(time=[2])
        >>> dr3.dataset['element_id'].values
        array([[  1.,   1.,  nan],
               [  3.,   3.,  nan],
               [ nan,   4.,  nan],
               [ nan,   4.,  nan]])
        >>> dr3.dataset['grid_element'].values
        array([['node', 'node', nan],
               ['link', 'link', nan],
               [nan, 'node', nan],
               [nan, 'node', nan]], dtype=object)

        And that forward filling fills everything as expected.

        >>> dr3.ffill_grid_element_and_id()
        >>> dr3.dataset['element_id'].values
        array([[  1.,   1.,   1.],
               [  3.,   3.,   3.],
               [ nan,   4.,   4.],
               [ nan,   4.,   4.]])

        >>> dr3.dataset['grid_element'].values
        array([['node', 'node', 'node'],
               ['link', 'link', 'link'],
               [nan, 'node', 'node'],
               [nan, 'node', 'node']], dtype=object)
        """

        ei = self._dataset["element_id"].values

        for i in range(ei.shape[0]):
            for j in range(1, ei.shape[1]):
                if np.isnan(ei[i, j]):
                    ei[i, j] = ei[i, j - 1]

        self._dataset["element_id"] = (["item_id", "time"], ei)

        ge = self._dataset["grid_element"].values
        for i in range(ge.shape[0]):
            for j in range(1, ge.shape[1]):
                if ge[i, j] not in self._permitted_locations:
                    ge[i, j] = ge[i, j - 1]
        self._dataset["grid_element"] = (["item_id", "time"], ge)

    @property
    def dataset(self):
        """The xarray Dataset that serves as the core datastructure."""
        return self._dataset

    @property
    def variable_names(self):
        """Return the name(s) of the data variable(s) in the record as a
        list."""
        _keys = []
        for key in self._dataset.to_dataframe().keys():
            _keys.append(key)
        return _keys

    @property
    def number_of_items(self):
        """Return the number of items in the DataRecord."""
        return len(self._dataset.item_id)

    @property
    def item_coordinates(self):
        """Return a list of the item_id coordinates in the DataRecord."""
        return self._dataset.item_id.values.tolist()

    @property
    def number_of_timesteps(self):
        """Return the number of time steps in the DataRecord."""
        return len(self._dataset.time)

    @property
    def time_coordinates(self):
        """Return a list of the time coordinates in the DataRecord."""
        return self._dataset.time.values.tolist()

    @property
    def earliest_time(self):
        """Return the earliest time coordinate in the DataRecord."""
        return min(self._dataset.time.values)

    @property
    def latest_time(self):
        """Return the latest time coordinate in the DataRecord."""
        return max(self._dataset.time.values)

    @property
    def prior_time(self):
        """Return the penultimate time coordinate in the DataRecord."""
        if self.number_of_timesteps < 2:
            return np.nan
        else:
            return sorted(self.time_coordinates)[-2]



#!/usr/bin/env python2
"""
Unit tests for KinwaveImplicitOverlandFlowModel.

Created on Sat Apr  1 10:49:33 2017

@author: gtucker
"""

from numpy.testing import assert_allclose, assert_raises

from landlab import RasterModelGrid
from landlab.components import LinearDiffusionOverlandFlowRouter


def test_steady_one_node():
    """Run to steady state with a single node"""
    xy_spacing = 2.0
    roughness = 0.01
    velocity_scale = 1.0
    rain_rate = 72.0 / (3600.0 * 1000.0)

    grid = RasterModelGrid((3, 3), xy_spacing=xy_spacing)
    grid.set_closed_boundaries_at_grid_edges(False, True, True, True)
    grid.add_zeros("topographic__elevation", at="node")
    olflow = LinearDiffusionOverlandFlowRouter(
        grid, rain_rate=rain_rate, roughness=roughness, velocity_scale=velocity_scale
    )
    for _ in range(18):
        olflow.run_one_step(20.0)
    actual = grid.at_node["surface_water__depth"][4]
    expected = (grid.dx * grid.dy * olflow.rain_rate / olflow.vel_coef) ** 0.3

    assert_allclose(actual, expected, atol=1e-4)


def test_invalid_parameters():
    """Make sure invalid parameters raise the right exceptions."""
    grid = RasterModelGrid((3, 3))
    grid.add_zeros("topographic__elevation", at="node")
    assert_raises(ValueError, LinearDiffusionOverlandFlowRouter, grid, roughness=-1.0)
    assert_raises(ValueError, LinearDiffusionOverlandFlowRouter, grid, rain_rate=-1.0)
    assert_raises(
        ValueError, LinearDiffusionOverlandFlowRouter, grid, infilt_depth_scale=0.0
    )
    assert_raises(ValueError, LinearDiffusionOverlandFlowRouter, grid, infilt_rate=-1.0)
    assert_raises(
        ValueError, LinearDiffusionOverlandFlowRouter, grid, velocity_scale=0.0
    )
    assert_raises(ValueError, LinearDiffusionOverlandFlowRouter, grid, cfl_factor=0.0)
    assert_raises(ValueError, LinearDiffusionOverlandFlowRouter, grid, cfl_factor=2.0)
    olflow = LinearDiffusionOverlandFlowRouter(grid)
    try:
        olflow.rain_rate = -1.0
    except ValueError:
        pass
    else:
        raise AssertionError("Setting negative rain_rate should raise ValueError")



cimport cython
cimport numpy as np

import numpy as np

DTYPE_INT = int
ctypedef np.int_t DTYPE_INT_t

DTYPE_FLOAT = np.float64
ctypedef np.float64_t DTYPE_FLOAT_t

@cython.boundscheck(False)
cpdef _landslide_runout(
    DTYPE_FLOAT_t dx,
    DTYPE_FLOAT_t phi,
    DTYPE_FLOAT_t min_deposition_slope,
    np.ndarray[DTYPE_INT_t, ndim=1] stack_rev_sel,
    np.ndarray[DTYPE_INT_t, ndim=2] receivers,
    np.ndarray[DTYPE_FLOAT_t, ndim=2] fract,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] Qs_in,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] L_Hill,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] Qs_out,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] dH_Hill,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] H_i_temp,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] max_D,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] length_adjacent_cells,
):
    """
    Calculate landslide runout using a non-local deposition algorithm, see:
        * Campforts B., Shobe C.M., Steer P., Vanmaercke M., Lague D., Braun J.
          (2020) HyLands 1.0: a hybrid landscape evolution model to simulate
          the impact of landslides and landslide-derived sediment on landscape
          evolution. Geosci Model Dev: 13(9):3863–86.
    """
    # define internal variables
    cdef int donor, rcvr, r
    cdef double accum, proportion, dH

    # Iterate backward through the stack, which means we work from upstream to
    # downstream.
    for donor in stack_rev_sel:
        dH = max(
                0,
                min(((Qs_in[donor] / dx) / L_Hill[donor]) / (1 - phi), max_D[donor])
            )
        dH_Hill[donor] += dH
        H_i_temp[donor] += dH

        Qs_in[donor] -= dH * dx * dx * (1 - phi)
        Qs_out[donor] += Qs_in[donor]

        for r in range(receivers.shape[1]):
            rcvr = receivers[donor, r]

            max_D_angle = H_i_temp[donor] - min_deposition_slope*length_adjacent_cells[r] - H_i_temp[rcvr]
            max_D[rcvr] = min(max(max_D[rcvr] , H_i_temp[donor] - H_i_temp[rcvr]),max_D_angle)

            proportion = fract[donor, r]
            if proportion > 0. and donor != rcvr:
                Qs_in[rcvr] += Qs_out[donor] * proportion
                Qs_in[donor] -= Qs_out[donor] * proportion



import numpy as np
from numpy.testing import assert_array_equal

from landlab import FramedVoronoiGrid


def test_rect_grid_link_order():
    """Test the order of links."""
    grid = FramedVoronoiGrid(
        (3, 2),
    )
    assert_array_equal(
        grid.nodes_at_link,
        [[0, 1], [0, 2], [2, 1], [1, 3], [2, 3], [2, 4], [4, 3], [3, 5], [4, 5]],
    )


def test_nodes_at_link():
    """Test nodes_at_link shares data with tail and head."""
    grid = FramedVoronoiGrid((3, 2))

    assert_array_equal(grid.nodes_at_link[:, 0], grid.node_at_link_tail)
    assert_array_equal(grid.nodes_at_link[:, 1], grid.node_at_link_head)

    assert np.may_share_memory(grid.nodes_at_link, grid.node_at_link_tail)
    assert np.may_share_memory(grid.nodes_at_link, grid.node_at_link_head)


def test_rect_face_at_link():
    grid = FramedVoronoiGrid((3, 3))
    assert_array_equal(
        grid.face_at_link,
        np.array([-1, -1, -1, 0, 1, 2, -1, 3, 4, -1, 5, 6, 7, -1, -1, -1]),
    )


def test_rect_link_at_face():
    grid = FramedVoronoiGrid((3, 3))
    assert_array_equal(grid.link_at_face, np.array([3, 4, 5, 7, 8, 10, 11, 12]))



#! /usr/bin/env python
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab.graph.framed_voronoi.framed_voronoi import HorizontalRectVoronoiGraph


@pytest.mark.parametrize(
    "n_cols", (1,) + tuple(np.random.randint(2, high=1000, size=5))
)
@pytest.mark.parametrize(
    "n_rows", (1,) + tuple(np.random.randint(2, high=1000, size=5))
)
def test_rect_perimeter_node_count(n_rows, n_cols, layout_graph):
    n_nodes = layout_graph.number_of_perimeter_nodes((n_rows, n_cols))

    assert n_nodes > 0
    assert n_nodes <= layout_graph.number_of_nodes((n_rows, n_cols))


def test_calc_perimeter_node_count_horizontal_rect():
    assert HorizontalRectVoronoiGraph.number_of_perimeter_nodes((1, 4)) == 4
    assert HorizontalRectVoronoiGraph.number_of_perimeter_nodes((3, 4)) == 10
    assert HorizontalRectVoronoiGraph.number_of_perimeter_nodes((4, 4)) == 12


@pytest.mark.parametrize("n_cols", [1, 2, 3, 7, 12])
@pytest.mark.parametrize("n_rows", [1, 2, 3, 7, 12])
def test_length_of_perimeter_nodes(layout_graph, n_rows, n_cols):
    assert len(
        layout_graph.perimeter_nodes((n_rows, n_cols))
    ) == layout_graph.number_of_perimeter_nodes((n_rows, n_cols))


def test_calc_perimeter_nodes_horizontal_rect():
    assert_array_equal(
        HorizontalRectVoronoiGraph.perimeter_nodes((3, 4)),
        (3, 7, 11, 10, 9, 8, 4, 0, 1, 2),
    )
    assert_array_equal(
        HorizontalRectVoronoiGraph.perimeter_nodes((4, 4)),
        (3, 7, 11, 15, 14, 13, 12, 8, 4, 0, 1, 2),
    )

    assert_array_equal(HorizontalRectVoronoiGraph.perimeter_nodes((1, 4)), (3, 2, 1, 0))
    assert_array_equal(HorizontalRectVoronoiGraph.perimeter_nodes((4, 1)), (0, 1, 2, 3))



import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal

import landlab.components.flow_director.cfuncs as _cfuncs
from landlab import HexModelGrid, RasterModelGrid


def test_hex():
    grid = HexModelGrid((3, 3))
    z = grid.x_of_node

    steepest_slope = np.zeros(grid.number_of_nodes)
    receiver = np.arange(grid.number_of_nodes)
    receiver_link = np.full(grid.number_of_nodes, -1, dtype=int)
    active_links = np.arange(grid.number_of_links, dtype=int)

    _cfuncs.adjust_flow_receivers(
        grid.node_at_link_head,
        grid.node_at_link_tail,
        z,
        grid.calc_grad_at_link(z),
        active_links,
        receiver,
        receiver_link,
        steepest_slope,
    )

    assert_array_equal(receiver, [3, 0, 1, 3, 3, 4, 5, 3, 7, 8])
    assert_array_equal(
        receiver_link,
        [2, 0, 1, -1, 8, 9, 10, 11, 17, 18],
    )
    assert_array_almost_equal(steepest_slope, [0.5, 1, 1, 0, 1, 1, 1, 0.5, 1, 1])


def test_raster_plane_dipping_north():
    grid = RasterModelGrid((3, 4))
    z = np.array(
        [
            [2.0, 2.0, 2.0, 2.0],
            [1.0, 1.0, 1.0, 1.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    ).flatten()

    steepest_slope = np.zeros(grid.number_of_nodes)
    receiver = np.arange(grid.number_of_nodes)
    receiver_link = np.full(grid.number_of_nodes, -1, dtype=int)
    active_links = np.arange(grid.number_of_links, dtype=int)

    _cfuncs.adjust_flow_receivers(
        grid.node_at_link_head,
        grid.node_at_link_tail,
        z,
        grid.calc_grad_at_link(z),
        active_links,
        receiver,
        receiver_link,
        steepest_slope,
    )

    assert_array_equal(
        receiver.reshape(grid.shape), [[4, 5, 6, 7], [8, 9, 10, 11], [8, 9, 10, 11]]
    )
    assert_array_equal(
        receiver_link.reshape(grid.shape),
        [[3, 4, 5, 6], [10, 11, 12, 13], [-1, -1, -1, -1]],
    )
    assert_array_almost_equal(
        steepest_slope.reshape(grid.shape), [[1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 0, 0]]
    )


def test_raster_plane_dipping_south():
    grid = RasterModelGrid((3, 4))
    z = np.array(
        [
            [0.0, 0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0, 1.0],
            [2.0, 2.0, 2.0, 2.0],
        ]
    ).flatten()

    steepest_slope = np.zeros(grid.number_of_nodes)
    receiver = np.arange(grid.number_of_nodes)
    receiver_link = np.full(grid.number_of_nodes, -1, dtype=int)
    active_links = np.arange(grid.number_of_links, dtype=int)

    _cfuncs.adjust_flow_receivers(
        grid.node_at_link_head,
        grid.node_at_link_tail,
        z,
        grid.calc_grad_at_link(z),
        active_links,
        receiver,
        receiver_link,
        steepest_slope,
    )

    assert_array_equal(
        receiver.reshape(grid.shape), [[0, 1, 2, 3], [0, 1, 2, 3], [4, 5, 6, 7]]
    )
    assert_array_equal(
        receiver_link.reshape(grid.shape),
        [[-1, -1, -1, -1], [3, 4, 5, 6], [10, 11, 12, 13]],
    )
    assert_array_almost_equal(
        steepest_slope.reshape(grid.shape), [[0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1]]
    )


def test_adjust_flow_receivers():
    # Grid of 5 nodes with these coordinates:
    # x = [0., 0., 3., 2., 3.]; y = [0., 4., 0., 2., 4.]
    # All nodes are linked (7 links)
    nodes_n = 5
    z = np.array([0.0, 1.0, 0.0, 4.0, 5.0])  # elevation
    receiver = -1 * np.ones(nodes_n, dtype=int)
    receiver_link = -1 * np.ones(nodes_n, dtype=int)
    steepest_slope = np.zeros(nodes_n, dtype=float)
    src_nodes = np.array([0, 0, 1, 0, 1, 3, 2, 3])  # tails
    dst_nodes = np.array([1, 2, 2, 3, 4, 2, 4, 4])  # heads
    link_slope = np.array(
        [-0.33333333, -0.0, 0.4472136, -1.0, -1.0]
        + [1.41421356, -2.23606798, -0.33333333]
    )

    _cfuncs.adjust_flow_receivers(
        src_nodes,
        dst_nodes,
        z,
        link_slope,
        np.arange(7),
        receiver,
        receiver_link,
        steepest_slope,
    )
    assert_array_equal(src_nodes, np.array([0, 0, 1, 0, 1, 3, 2, 3]))
    assert_array_equal(dst_nodes, np.array([1, 2, 2, 3, 4, 2, 4, 4]))
    assert_array_equal(receiver, np.array([-1, 2, -1, 2, 2]))
    assert_array_equal(receiver_link, np.array([-1, 2, -1, 5, 6]))
    assert_array_almost_equal(
        steepest_slope, np.array([0.0, 0.4472136, 0.0, 1.41421356, 2.23606798])
    )



#! /usr/bin/env python
import argparse
import os


def requirements(extras, include_required=False):
    sections = Requirements().get_sections(
        extras, include_required=not extras or include_required
    )

    print("# Requirements extracted from pyproject.toml")
    for section, packages in sections.items():
        print(f"# {section}")
        print(os.linesep.join(sorted(packages)))


class Requirements:
    def __init__(self):
        tomllib = _find_tomllib()

        with open("pyproject.toml", "rb") as fp:
            project = tomllib.load(fp)["project"]
        self._dependencies = Requirements._clean_dependency_names(
            project.get("dependencies", [])
        )
        self._optional_dependencies = {
            name: Requirements._clean_dependency_names(packages)
            for name, packages in project.get("optional-dependencies", {}).items()
        }

    @staticmethod
    def _clean_dependency_names(packages):
        return [package.split(";")[0].strip() for package in packages]

    @property
    def required(self):
        return tuple(self._dependencies)

    def get_optional(self, extra):
        return tuple(self._optional_dependencies[extra])

    def get_sections(self, sections=None, include_required=False):
        sections = sections or []
        reqs = {}
        if include_required:
            reqs["[project.dependencies]"] = self.required
        for section in sections:
            reqs[f"[project.optional-dependencies] {section}"] = self.get_optional(
                section
            )
        return reqs


def _find_tomllib():
    try:
        import tomllib
    except ModuleNotFoundError:
        import tomli as tomllib
    return tomllib


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Extract requirements information from pyproject.toml"
    )
    parser.add_argument("extras", type=str, nargs="*")
    args = parser.parse_args()

    requirements(args.extras)



#! /usr/env/python
"""Python implementation of RadialModelGrid, a grid class used to create and
manage structured Voronoi-Delaunay grids for 2D numerical models.

Do NOT add new documentation here. Grid documentation is now built in a
semi- automated fashion. To modify the text seen on the web, edit the
files `docs/text_for_[gridfile].py.txt`.
"""

import numpy as np
import xarray as xr

from ..graph import DualRadialGraph
from .base import ModelGrid


class RadialModelGrid(DualRadialGraph, ModelGrid):

    """Grid of concentric circles.

    This inherited class implements a circular grid in which grid nodes are
    placed at regular radial and semi-regular arc-wise intervals. That is,
    if the radial spacing between *shells* is *dr*, the nodes are placed around
    the circular shell at regular intervals that get as close as possible to
    *dr*. The points are then arranged in a Delaunay triangulation with Voronoi
    cells. Within each ring, nodes are numbered according to Landlab
    convention, from the first node counterclockwise of east. Numbering
    begins at the centermost node and works outwards through the rings.
    """

    def __init__(
        self,
        n_rings=0,
        nodes_in_first_ring=6,
        spacing=1.0,
        xy_of_center=(0.0, 0.0),
        xy_of_reference=(0.0, 0.0),
        xy_axis_name=("x", "y"),
        xy_axis_units="-",
    ):
        """Create a circular grid.

        Create a circular grid in which grid nodes are placed at regular
        radial and semi-regular arc-wise intervals. That is, if the radial
        spacing between *shells* is *dr*, the nodes are placed around the
        circular shell at regular intervals that get as close as possible to
        *dr*.  The points are then arranged in a Delaunay triangulation with
        Voronoi cells.


        Parameters
        ----------
        n_rings : int
            Number of rings in the grid.
        nodes_in_first_ring : int, optional
            Number of nodes in the first ring.
        spacing : float, optional
            Distance between rings.
        xy_of_center : tuple, optional
            (x, y) coordinate of center point. Default
            is (0., 0.)
        xy_of_reference : tuple, optional
            Coordinate value in projected space of the reference point,
            `xy_of_lower_left`. Default is (0., 0.)

        Returns
        -------
        RadialModelGrid
            A newly-created grid.

        Examples
        --------
        A grid with just one ring will have a node at the origin surrounded
        by six other nodes by default. This can be changed by providing the
        keyword argument `nodes_in_first_ring`.

        >>> import numpy as np
        >>> from landlab import RadialModelGrid
        >>> omg = RadialModelGrid(
        ...     n_rings=1, nodes_in_first_ring=8, xy_of_center=(0., 0.)
        ... )
        >>> omg.number_of_nodes
        9
        >>> omg.number_of_cells
        1

        A second rings will have 16 nodes (1 + 8 + 16 = 25).

        >>> omg = RadialModelGrid(2, nodes_in_first_ring=8)
        >>> omg.number_of_nodes
        25

        >>> np.round(omg.radius_at_node)
        array([ 2.,  2.,  2.,  2.,  2.,  1.,  2.,  2.,  1.,  1.,  2.,  1.,  0.,
                1.,  2.,  1.,  1.,  2.,  2.,  1.,  2.,  2.,  2.,  2.,  2.])
        """
        xy_of_center = tuple(xy_of_center)

        DualRadialGraph.__init__(
            self,
            (n_rings, nodes_in_first_ring),
            spacing=spacing,
            xy_of_center=xy_of_center,
            sort=True,
        )
        ModelGrid.__init__(
            self,
            xy_of_reference=xy_of_reference,
            xy_axis_name=xy_axis_name,
            xy_axis_units=xy_axis_units,
        )

        self._node_status = np.full(
            self.number_of_nodes, self.BC_NODE_IS_CORE, dtype=np.uint8
        )
        self._node_status[self.perimeter_nodes] = self.BC_NODE_IS_FIXED_VALUE
        self._xy_of_center = tuple(xy_of_center)

    @classmethod
    def from_dict(cls, kwds):
        args = ()
        return cls(*args, **kwds)

    @classmethod
    def from_dataset(cls, dataset):
        return cls(
            n_rings=int(dataset["n_rings"]),
            nodes_in_first_ring=int(dataset["nodes_in_first_ring"]),
            spacing=float(dataset["spacing"]),
            xy_of_center=dataset["xy_of_center"],
        )

    def as_dataset(self, include="*", exclude=None, time=None):
        dataset = xr.Dataset(
            {
                "n_rings": self.number_of_rings,
                "nodes_in_first_ring": self.number_of_nodes_in_ring[0],
                "spacing": self.spacing_of_rings,
                "xy_of_center": (("dim",), list(self.xy_of_center)),
            },
            attrs={"grid_type": "radial"},
        )
        return dataset.update(
            super().as_dataset(include=include, exclude=exclude, time=time)
        )

    @property
    def xy_of_center(self):
        """Return (x, y) of the reference point."""
        return self._xy_of_center

    @xy_of_center.setter
    def xy_of_center(self, xy_of_center):
        """Set a new value for the xy_of_lower_left."""
        dx = self.xy_of_center[0] - xy_of_center[0]
        dy = self.xy_of_center[1] - xy_of_center[1]
        with self.thawed():
            self.x_of_node[:] -= dx
            self.y_of_node[:] -= dy
        self._xy_of_center = xy_of_center



import hypothesis.extra.numpy as hynp
import numpy as np
import pytest
from hypothesis import given
from hypothesis.strategies import composite, floats, integers, lists
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.grid.create_network import (
    AlongChannelSpacingAtLeast,
    AtMostNodes,
    ChannelSegment,
    ChannelSegmentConnector,
    JustEndNodes,
    SegmentLinkCollector,
    SegmentNodeCoordinateCollector,
    SegmentNodeReindexer,
    SpacingAtLeast,
    _reduce_nodes,
    _reduce_to_fewest_nodes,
    create_network_links,
    create_xy_of_node,
    get_node_fields,
    network_grid_from_raster,
    network_grid_from_segments,
    pairwise,
    reindex_network_nodes,
    spacing_from_drainage_area,
)


@given(
    drainage_area=hynp.arrays(
        dtype=hynp.floating_dtypes(),
        shape=hynp.array_shapes(),
        elements=floats(min_value=0, width=16),
    )
)
def test_calc_spacing_always_positive(drainage_area):
    assert np.all(spacing_from_drainage_area(drainage_area) >= 0.0)


@given(
    drainage_area=hynp.arrays(
        dtype=hynp.floating_dtypes(),
        shape=hynp.array_shapes(),
        elements=floats(min_value=0, width=16),
    )
)
def test_calc_spacing_unit_keywords(drainage_area):
    spacing = spacing_from_drainage_area(drainage_area, a=1, b=1, n_widths=1)

    assert np.allclose(spacing, drainage_area / 1e6)


@given(nodes=lists(integers(), min_size=2, max_size=1024))
def test_channel_segment(nodes):
    segment = ChannelSegment(nodes)
    assert segment.downstream_node == nodes[0]
    assert segment.upstream_node == nodes[-1]
    assert_array_equal(segment.nodes, nodes)
    assert len(segment) == len(nodes)


@given(nodes=lists(integers(), min_size=2, max_size=1024))
def test_channel_segment_set_nodes(nodes):
    segment = ChannelSegment([0, 1])

    segment.nodes = nodes
    assert segment.downstream_node == nodes[0]
    assert segment.upstream_node == nodes[-1]
    assert_array_equal(segment.nodes, nodes)
    assert len(segment) == len(nodes)


def test_channel_segment_add_downstream_node():
    segment = ChannelSegment([0, 1])
    downstream = ChannelSegment([5, 6])
    assert segment.downstream is None
    assert len(downstream.upstream) == 0

    segment.downstream = downstream

    assert segment.downstream is downstream
    assert segment in downstream.upstream


def test_channel_segment_add_upstream_node():
    segment = ChannelSegment([0, 1])
    upstream = ChannelSegment([5, 6])
    assert len(segment.upstream) == 0
    assert upstream.downstream is None

    segment.add_upstream(upstream)

    assert upstream in segment.upstream
    assert upstream.downstream is segment


@given(segments=lists(lists(integers(), min_size=2, max_size=1024), min_size=1))
def test_channel_segment_many_upstream(segments):
    segments = [ChannelSegment(segment) for segment in segments]
    root = segments[0]
    for current, next in pairwise(segments):
        current.add_upstream(next)

    assert root.count_segments(direction="upstream") == len(segments) - 1
    assert root.count_segments(direction="downstream") == 0


@given(segments=lists(lists(integers(), min_size=2, max_size=1024), min_size=1))
def test_channel_segment_many_flat_upstream(segments):
    segments = [ChannelSegment(segment) for segment in segments]
    root = segments[0]
    for segment in segments[1:]:
        root.add_upstream(segment)
    assert root.downstream is None
    assert len(root.upstream) == len(segments) - 1

    assert root.count_segments(direction="upstream") == len(segments) - 1
    assert root.count_segments(direction="downstream") == 0


@given(segments=lists(lists(integers(), min_size=2, max_size=1024), min_size=1))
def test_channel_segment_many_downstream(segments):
    segments = [ChannelSegment(segment) for segment in segments]
    root = segments[0]
    for current, next in pairwise(segments):
        current.downstream = next
    root = segments[0]
    leaf = segments[-1]

    assert root.count_segments(direction="upstream") == 0
    assert root.count_segments(direction="downstream") == len(segments) - 1

    assert leaf.count_segments(direction="upstream") == len(segments) - 1
    assert leaf.count_segments(direction="downstream") == 0


@given(nodes=lists(integers(), min_size=2, max_size=1024))
def test_channel_segment_for_each(nodes):
    all_nodes = []

    def collect_nodes(segment):
        all_nodes.extend(list(segment.nodes))

    segment = ChannelSegment(nodes)
    segment.for_each(collect_nodes)

    assert_array_equal(all_nodes, segment.nodes)


def test_connector_add_upstream():
    segment = ChannelSegment([0, 1])
    connector = ChannelSegmentConnector(segment)

    assert connector.root is segment
    assert len(connector.orphans) == 0

    connector.add(ChannelSegment([1, 2]))
    assert connector.root is segment
    assert connector.root.count_segments(direction="upstream") == 1
    assert connector.root.downstream is None


def test_connector_add_downstream():
    segment_1 = ChannelSegment([0, 1])
    segment_2 = ChannelSegment([2, 0])
    connector = ChannelSegmentConnector(segment_1)

    connector.add(segment_2)
    assert connector.root is segment_2
    assert connector.root.count_segments(direction="upstream") == 1
    assert connector.root.downstream is None


def test_connector_add_orphan():
    segment_1 = ChannelSegment([0, 1])
    segment_2 = ChannelSegment([2, 3])
    connector = ChannelSegmentConnector(segment_1)

    connector.add(segment_2)

    assert connector.root is segment_1
    assert connector.root.count_segments(direction="upstream") == 0
    assert connector.root.downstream is None

    assert len(connector.orphans) == 1
    assert connector.orphans == (segment_2,)

    connector.add(ChannelSegment([1, 2]))
    assert connector.root.count_segments(direction="upstream") == 2
    assert connector.orphans == ()


_grid_dims_to_test = integers(min_value=3, max_value=128)


@composite
def shape_and_indices(draw, elements=_grid_dims_to_test):
    shape = draw(lists(elements, min_size=2, max_size=2))
    indices = draw(
        lists(
            integers(min_value=0, max_value=shape[0] * shape[1] - 1),
            min_size=1,
            max_size=1024,
        )
    )
    return shape, indices


@given(shape_and_segment=shape_and_indices())
def test_construct_xy_of_node(shape_and_segment):
    shape, segment = shape_and_segment
    grid = RasterModelGrid(shape)

    collect_coordinates = SegmentNodeCoordinateCollector(grid)
    collect_coordinates(ChannelSegment(segment))
    xy_of_node = collect_coordinates.xy_of_node

    assert len(xy_of_node) == len(segment)

    x_of_node, y_of_node = zip(*xy_of_node)
    assert (x_of_node == grid.x_of_node[segment]).all()
    assert (y_of_node == grid.y_of_node[segment]).all()


@given(nodes=lists(integers(), min_size=0, max_size=1024))
def test_reindex_segment_nodes_orphan(nodes):
    segment = ChannelSegment(nodes)
    reindex = SegmentNodeReindexer()
    reindex(segment)
    assert segment.nodes == list(range(len(nodes)))


@given(nodes=lists(integers(), min_size=0, max_size=1024), last_node=integers())
def test_reindex_segment_nodes_with_last_node(nodes, last_node):
    segment = ChannelSegment(nodes)
    reindex = SegmentNodeReindexer(nodes=[last_node])
    reindex(segment)
    assert segment.nodes == list(range(last_node + 1, last_node + 1 + len(nodes)))


@given(nodes=lists(integers(), min_size=0, max_size=1024), last_node=integers())
def test_reindex_segment_nodes_with_downstream(nodes, last_node):
    root = ChannelSegment([0, 1])
    segment = ChannelSegment(nodes)
    root.add_upstream(segment)

    reindex = SegmentNodeReindexer(nodes=[last_node])
    reindex(segment)

    assert segment.nodes[0] == root.nodes[-1]
    assert segment.nodes[1:] == list(range(last_node + 1, last_node + len(nodes)))


@given(nodes=lists(integers(), min_size=2, max_size=1024))
def test_create_links(nodes):
    segment = ChannelSegment(nodes)

    collect_links = SegmentLinkCollector()
    collect_links(segment)
    links = collect_links.links

    assert len(links) == len(segment) - 1
    heads, tails = zip(*links)
    assert list(heads) == nodes[:-1]
    assert list(tails) == nodes[1:]


@given(nodes=lists(integers(), min_size=2, max_size=1024))
def test_create_links_with_existing(nodes):
    segment = ChannelSegment(nodes)

    collect_links = SegmentLinkCollector(links=[(1, 2), (3, 4)])
    collect_links(segment)
    links = collect_links.links

    assert links[:2] == [(1, 2), (3, 4)]
    assert len(links[2:]) == len(segment) - 1
    heads, tails = zip(*links[2:])
    assert list(heads) == nodes[:-1]
    assert list(tails) == nodes[1:]


@given(nodes=lists(integers(), min_size=2, max_size=1024))
def test_create_links_with_downstream(nodes):
    root = ChannelSegment([0, 1])
    segment = ChannelSegment(nodes)
    root.add_upstream(segment)

    collect_links = SegmentLinkCollector()
    collect_links(segment)
    links = collect_links.links

    assert len(links) == len(segment) - 1
    assert links[0] == (root.nodes[-1], segment.nodes[1])

    if len(links) > 1:
        heads, tails = zip(*links[1:])
        assert list(heads) == nodes[1:-1]
        assert list(tails) == nodes[2:]


def test_reindex_network_nodes():
    root = ChannelSegmentConnector([10, 11, 12], [12, 13], [12, 14], [14, 15]).root
    reindex_network_nodes(root)
    assert list(root.nodes) == [0, 1, 2]
    assert list(root.upstream[0].nodes) == [2, 3]
    assert list(root.upstream[1].nodes) == [2, 4]
    assert list(root.upstream[1].upstream[0].nodes) == [4, 5]


def test_create_network_links():
    root = ChannelSegmentConnector([0, 1, 2], [2, 3], [2, 4], [4, 5]).root
    links = create_network_links(root)
    assert links == [(0, 1), (1, 2), (2, 3), (2, 4), (4, 5)]


def test_graph_from_segments():
    r"""
    ::
      *
      |
      *   *
       \ /
    *   *
    |   |
    *   *     *
    |   |    /
    *   * - *
     \ /
      *
       \
        *
        |
        *
    """
    grid = RasterModelGrid((8, 6))
    grid.at_node["z"] = list(range(grid.number_of_nodes))
    segments = [
        [3, 9, 14],
        [14, 19, 25, 31],
        [14, 21],
        [21, 27, 33],
        [33, 40],
        [33, 38, 44],
        [21, 22, 29],
    ]
    graph = network_grid_from_segments(grid, segments)
    assert graph.number_of_nodes == 14
    assert graph.number_of_links == 13

    assert list(zip(graph.x_of_node, graph.y_of_node)) == [
        (3.0, 0.0),
        (3.0, 1.0),
        (2.0, 2.0),
        (1.0, 3.0),
        (3.0, 3.0),
        (4.0, 3.0),
        (1.0, 4.0),
        (3.0, 4.0),
        (5.0, 4.0),
        (1.0, 5.0),
        (3.0, 5.0),
        (2.0, 6.0),
        (4.0, 6.0),
        (2.0, 7.0),
    ]

    assert "z" in graph.at_node
    assert list(graph.at_node["z"]) == [
        3,
        9,
        14,
        19,
        21,
        22,
        25,
        27,
        29,
        31,
        33,
        38,
        40,
        44,
    ]


def test_reduce_nodes():
    nodes = _reduce_nodes([0.0, 1.0, 2.0, 3.0, 3.5], spacing=1.0)
    assert nodes == [0, 1, 2, 3, 4]

    nodes = _reduce_nodes([0.0, 1.0, 2.0, 3.0, 3.5], spacing=0.5)
    assert nodes == [0, 1, 2, 3, 4]

    nodes = _reduce_nodes([0.0, 1.0, 2.0, 3.0, 4.0], spacing=1.75)
    assert nodes == [0, 2, 4]

    nodes = _reduce_nodes([0.0, 1.0, 2.0, 3.0, 4.0], spacing=[1.0, 1.0, 2.0, 2.0, 2.0])
    assert nodes == [0, 1, 2, 4]

    nodes = _reduce_nodes([0.0, 1.0, 2.0, 3.0, 4.0], spacing=1000.0)
    assert nodes == [0, 4]

    nodes = _reduce_nodes([0.0, 1.0, 2.0, 3.0, 4.0, 5.0], spacing=2.0)
    assert nodes == [0, 2, 4, 5]


def test_reduce_to_fewest_nodes():
    x = [0.0, 1.0, 2.0, 3.0, 3.5]
    y = [0.0] * len(x)
    nodes = _reduce_to_fewest_nodes(list(zip(x, y)), spacing=1.0)
    assert nodes == [0, 1, 2, 3, 4]

    nodes = _reduce_to_fewest_nodes(list(zip(x, y)), spacing=0.5)
    assert nodes == [0, 1, 2, 3, 4]

    x = [0.0, 1.0, 2.0, 3.0, 4.0]
    y = [0.0] * len(x)
    nodes = _reduce_to_fewest_nodes(list(zip(x, y)), spacing=1.75)
    assert nodes == [0, 2, 4]

    nodes = _reduce_to_fewest_nodes(list(zip(x, y)), spacing=[1.0, 1.0, 2.0, 2.0, 2.0])
    assert nodes == [0, 1, 2, 4]

    nodes = _reduce_to_fewest_nodes(list(zip(x, y)), spacing=1000.0)
    assert nodes == [0, 4]

    x = [0.0, 1.0, 2.0, 3.0, 4.0, 5.0]
    y = [0.0] * len(x)
    nodes = _reduce_to_fewest_nodes(list(zip(x, y)), spacing=2.0)
    assert nodes == [0, 2, 4, 5]


def test_reduce_nodes_stay_the_same():
    nodes = _reduce_nodes([0.0, 1.0, 2.0, 3.0, 4.0], spacing=1.0)
    assert nodes == [0, 1, 2, 3, 4]

    nodes = _reduce_nodes([0.0, 2.0, 4.0, 6.0, 8.0], spacing=2.0)
    assert nodes == [0, 1, 2, 3, 4]

    nodes = _reduce_nodes([0.0, 1.0, 2.0, 3.0, 4.0], spacing=0.5)
    assert nodes == [0, 1, 2, 3, 4]

    nodes = _reduce_nodes([0.0, 1.0, 3.0, 6.0, 10.0], spacing=[1, 2, 3, 4, 5])
    assert nodes == [0, 1, 2, 3, 4]


@pytest.mark.parametrize(
    "x,spacing",
    [
        ([0.0, 1.0, 2.0, 3.0, 4.0], 1.0),
        ([0.0, 2.0, 4.0, 6.0, 8.0], 2.0),
        ([0.0, 1.0, 2.0, 3.0, 4.0], 0.5),
        ([0.0, 1.0, 3.0, 6.0, 10.0], [1, 2, 3, 4, 5]),
    ],
)
def test_reduce_to_fewest_nodes_stay_the_same(x, spacing):
    y = [0.0] * len(x)
    nodes = _reduce_to_fewest_nodes(list(zip(x, y)), spacing=spacing)
    assert nodes == [0, 1, 2, 3, 4]


@given(
    spacing=hynp.arrays(
        dtype=float,
        shape=hynp.array_shapes(min_dims=1, max_dims=1, min_side=2),
        elements=floats(min_value=1e-3, max_value=1e3),
    )
)
def test_reduce_nodes_min_max_spacing(spacing):
    distance_along_segment = np.cumsum(spacing)

    if np.any(np.diff(distance_along_segment) <= 0):
        raise ValueError(f"array not sorted ({distance_along_segment})")

    nodes = _reduce_nodes(distance_along_segment, spacing=spacing.min())
    assert np.all(nodes == np.arange(len(spacing)))

    nodes = _reduce_nodes(
        distance_along_segment,
        spacing=distance_along_segment[-1] - distance_along_segment[0],
    )
    assert nodes == [0, len(spacing) - 1]


@given(
    spacing=hynp.arrays(
        dtype=float,
        shape=hynp.array_shapes(min_dims=1, max_dims=1, min_side=2),
        elements=floats(min_value=1e-3, max_value=1e3),
    )
)
def test_reduce_to_fewest_nodes_min_max_spacing(spacing):
    distance_along_segment = np.cumsum(spacing)

    if np.any(np.diff(distance_along_segment) <= 0):
        raise ValueError(f"array not sorted ({distance_along_segment})")

    xy_of_node = list(zip(distance_along_segment, [0.0] * len(distance_along_segment)))
    min_spacing = np.diff(distance_along_segment).min()

    nodes = _reduce_to_fewest_nodes(xy_of_node, spacing=min_spacing)
    assert np.all(nodes == np.arange(len(spacing)))

    nodes = _reduce_to_fewest_nodes(
        xy_of_node,
        spacing=distance_along_segment[-1] - distance_along_segment[0],
    )
    assert nodes == [0, len(spacing) - 1]


def test_educe_to_fewest_nodes_wraparound():
    x = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]
    y = [0.0, 1.0, 2.0, 2.0, 1.0, 0.0]

    assert _reduce_to_fewest_nodes(list(zip(x, y)), spacing=1.001) == [0, 5]


def test_create_xy_of_node_with_branch():
    grid = RasterModelGrid((3, 4), xy_spacing=(2.0, 3.0))
    network = ChannelSegmentConnector([4, 5], [5, 2, 3], [5, 10, 11])
    xy_of_node = create_xy_of_node(network.root, grid)

    assert np.allclose(xy_of_node, [[0, 3], [2, 3], [4, 0], [6, 0], [4, 6], [6, 6]])


@given(
    nodes=lists(
        integers(min_value=0, max_value=1023),
        min_size=2,
        max_size=1024,
    ),
)
def test_create_xy_of_node_one_segement(nodes):
    grid = RasterModelGrid((16, 64), xy_spacing=(2.0, 3.0))
    network = ChannelSegmentConnector(nodes)
    xy_of_node = create_xy_of_node(network.root, grid)

    assert np.allclose(xy_of_node[:, 0], grid.x_of_node[nodes])
    assert np.allclose(xy_of_node[:, 1], grid.y_of_node[nodes])


def test_xy_of_node_if_not_network_root():
    grid = RasterModelGrid((3, 4), xy_spacing=(2.0, 3.0))
    network = ChannelSegmentConnector([4, 5], [5, 2], [5, 10, 11], [2, 3], [2, 7])

    base = network.root.upstream[0]
    xy_of_node = create_xy_of_node(base, grid)

    assert np.allclose(xy_of_node, [[4, 0], [6, 0], [6, 3]])


def test_get_node_fields_one_field():
    grid = RasterModelGrid((3, 4))
    grid.at_node["foo"] = np.arange(12) * 10
    network = ChannelSegmentConnector([0, 5], [5, 6, 7], [5, 9])

    fields = get_node_fields(network.root, grid)
    assert list(fields) == ["foo"]
    assert_array_equal(fields["foo"], [0, 50, 60, 70, 90])


def test_get_node_fields_two_fields():
    grid = RasterModelGrid((3, 4))
    grid.at_node["foo"] = np.arange(12) * 10
    grid.at_node["bar"] = np.arange(12) * 100

    network = ChannelSegmentConnector([0, 5], [5, 6, 7], [5, 9])

    fields = get_node_fields(network.root, grid)
    assert sorted(fields) == ["bar", "foo"]
    assert_array_equal(fields["foo"], [0, 50, 60, 70, 90])
    assert_array_equal(fields["bar"], [0, 500, 600, 700, 900])


def test_get_node_fields_include():
    grid = RasterModelGrid((3, 4))
    grid.at_node["foo"] = np.arange(12) * 10
    grid.at_node["bar"] = np.arange(12) * 100
    grid.at_node["baz"] = np.arange(12) * 1000

    network = ChannelSegmentConnector([0, 5], [5, 6, 7], [5, 9])

    fields = get_node_fields(network.root, grid, include="at_node:f*")
    assert list(fields) == ["foo"]
    assert_array_equal(fields["foo"], [0, 50, 60, 70, 90])

    fields = get_node_fields(network.root, grid, include="at_node:b*")
    assert sorted(fields) == ["bar", "baz"]
    assert_array_equal(fields["bar"], [0, 500, 600, 700, 900])
    assert_array_equal(fields["baz"], [0, 5000, 6000, 7000, 9000])


def test_get_node_fields_exclude():
    grid = RasterModelGrid((3, 4))
    grid.add_empty("foo", at="node")
    grid.add_empty("bar", at="node")
    grid.add_empty("baz", at="node")

    network = ChannelSegmentConnector([0, 5], [5, 6, 7], [5, 9])

    expected = get_node_fields(network.root, grid, include="at_node:b*")
    actual = get_node_fields(network.root, grid, exclude="at_node:f*")

    assert actual.keys() == expected.keys()
    for name in actual:
        assert_array_equal(actual[name], expected[name])


def test_get_node_fields_ignore_non_node_fields():
    grid = RasterModelGrid((3, 4))
    grid.add_empty("foo", at="node")
    grid.add_empty("bar", at="node")
    grid.add_empty("baz", at="link")

    network = ChannelSegmentConnector([0, 5], [5, 6, 7], [5, 9])

    fields = get_node_fields(network.root, grid, include="*")

    assert sorted(fields) == ["bar", "foo"]


def test_network_grid_from_raster():
    grid = RasterModelGrid((4, 5))
    grid.at_node["topographic__elevation"] = np.flipud(
        np.asarray(
            [
                [4, 4, 4, 4, 4],
                [4, 4, 2, 4, 4],
                [4, 4, 1, 4, 4],
                [4, 4, 0, 4, 4],
            ],
            dtype=float,
        )
    )

    network = network_grid_from_raster(grid)
    assert network.number_of_nodes == 7
    assert network.number_of_links == 6
    assert np.allclose(
        network.xy_of_node,
        [
            [2.0, 0.0],
            [1.0, 1.0],
            [2.0, 1.0],
            [3.0, 1.0],
            [1.0, 2.0],
            [2.0, 2.0],
            [3.0, 2.0],
        ],
    )
    assert_array_equal(
        network.nodes_at_link, [[0, 2], [1, 2], [2, 3], [4, 2], [2, 5], [2, 6]]
    )


@given(nodes=lists(integers(), min_size=2))
def test_reducer_just_end_nodes(nodes):
    reduce = JustEndNodes()
    assert_array_equal(reduce(nodes), [nodes[0], nodes[-1]])


@given(nodes=lists(integers(), min_size=3))
def test_reducer_min_three_nodes(nodes):
    reduce = AtMostNodes()
    assert_array_equal(reduce(nodes), [nodes[0], nodes[len(nodes) // 2], nodes[-1]])


@given(nodes=lists(integers(), min_size=2))
def test_reducer_min_ndoes_matches_just_end_nodes(nodes):
    just_end_nodes = JustEndNodes()
    at_most_two_nodes = AtMostNodes(count=2)
    assert_array_equal(just_end_nodes(nodes), at_most_two_nodes(nodes))


@given(nodes=lists(integers(), min_size=2))
def test_reducer_min_nodes_no_change(nodes):
    reduce = AtMostNodes(count=len(nodes) + 1)
    assert_array_equal(reduce(nodes), nodes)


@pytest.mark.parametrize("count", [-1, 0, 1])
def test_reducer_min_nodes_less_than_two(count):
    with pytest.raises(ValueError):
        AtMostNodes(count=count)


def test_reducer_spacing_at_least():
    grid = RasterModelGrid((3, 6), xy_spacing=(3.0, 4.0))
    reduce = SpacingAtLeast(xy_of_node=grid.xy_of_node)

    assert_array_equal(
        reduce.calc_distance_along_segment([6, 7, 8, 9, 10, 11]), [0, 3, 6, 9, 12, 15]
    )
    assert_array_equal(reduce.calc_distance_along_segment([0, 7, 14]), [0, 5, 10])
    assert_array_equal(reduce.calc_distance_along_segment([0, 6, 12]), [0, 4, 8])

    reduce = SpacingAtLeast(xy_of_node=grid.xy_of_node, spacing=3.0)
    assert_array_equal(reduce([6, 7, 8, 9, 10, 11]), [6, 7, 8, 9, 10, 11])

    reduce = SpacingAtLeast(xy_of_node=grid.xy_of_node, spacing=1.5)
    assert_array_equal(reduce([6, 7, 8, 9, 10, 11]), [6, 7, 8, 9, 10, 11])

    reduce = SpacingAtLeast(xy_of_node=grid.xy_of_node, spacing=6.0)

    assert_array_equal(reduce([6, 7, 8, 9, 10, 11]), [6, 8, 10, 11])


def test_reducer_spacing_at_least_variable():
    xy_of_node = [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0], [5, 0]]
    spacing = [1, 2, 3, 4, 5, 6]

    reduce = SpacingAtLeast(xy_of_node=xy_of_node, spacing=spacing)

    assert_array_equal(reduce([0, 1, 2, 3, 4, 5]), [0, 1, 3, 5])
    assert_array_equal(reduce([0, 1, 2, 3, 4, 5]), reduce(reduce([0, 1, 2, 3, 4, 5])))


@given(
    xy_of_node=hynp.arrays(
        dtype=float,
        shape=hynp.array_shapes(min_dims=1, max_dims=1, min_side=4, max_side=128),
        elements=integers(min_value=-1024, max_value=1024),
        unique=True,
    ),
    spacing=floats(min_value=0.0, exclude_min=True),
)
def test_reducer_spacing_at_least_all_greater(xy_of_node, spacing):
    xy_of_node = xy_of_node[: len(xy_of_node) - len(xy_of_node) % 2].reshape((-1, 2))
    xy_of_node /= 100.0

    segment = np.arange(len(xy_of_node))
    reduce = SpacingAtLeast(xy_of_node=xy_of_node, spacing=spacing)
    reduced_segment = reduce(segment)

    distance_along_segment = reduce.calc_distance_along_segment(reduced_segment[:-1])

    assert reduced_segment[0] == segment[0]
    assert reduced_segment[-1] == segment[-1]
    assert np.all(np.diff(distance_along_segment) >= spacing)


def test_reducer_along_channel_spacing_at_least_variable():
    xy_of_node = [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0], [5, 0]]
    spacing = [1, 2, 3, 4, 5, 6]

    reduce = AlongChannelSpacingAtLeast(xy_of_node=xy_of_node, spacing=spacing)

    assert_array_equal(reduce([0, 1, 2, 3, 4, 5]), [0, 1, 3, 5])
    assert_array_equal(reduce([0, 1, 2, 3, 4, 5]), reduce(reduce([0, 1, 2, 3, 4, 5])))



import pytest

from landlab import RasterModelGrid
from landlab.components import LandslideProbability


@pytest.fixture
def ls_prob():
    grid = RasterModelGrid((20, 20), xy_spacing=10e0)
    grid.add_zeros("topographic__slope", at="node", dtype=float)
    grid.add_zeros("topographic__specific_contributing_area", at="node")
    grid.add_zeros("soil__transmissivity", at="node")
    grid.add_zeros("soil__saturated_hydraulic_conductivity", at="node")
    grid.add_zeros("soil__mode_total_cohesion", at="node")
    grid.add_zeros("soil__minimum_total_cohesion", at="node")
    grid.add_zeros("soil__maximum_total_cohesion", at="node")
    grid.add_zeros("soil__internal_friction_angle", at="node")
    grid.add_zeros("soil__density", at="node")
    grid.add_zeros("soil__thickness", at="node")

    return LandslideProbability(grid)



"""Accumulate flow and calc drainage area, while permitting gain or loss
of discharge during flow.

DEJH, late 2018
"""

import sys

from landlab.components.flow_accum import (
    FlowAccumulator,
    flow_accum_bw,
    flow_accum_to_n,
)

if sys.version_info[0] >= 3:
    from inspect import signature


class LossyFlowAccumulator(FlowAccumulator):

    """Component to calculate drainage area and accumulate flow, while
    permitting dynamic loss or gain of flow downstream.

    This component is closely related to the :class:`.FlowAccumulator`,
    in that this is accomplished by first finding flow directions by a user-specified
    method and then calculating the drainage area and discharge. However,
    this component additionally requires the passing of a function that
    describes how discharge is lost or gained downstream::

        f(Qw, nodeID, linkID, grid)

    See the Examples below to see how this works in practice.

    Optionally, spatially variable runoff can be set either by the model grid
    field ``"water__unit_flux_in"`` or the input variable ``runoff_rate``.

    Optionally a depression finding component can be specified and flow
    directing, depression finding, and flow routing can all be accomplished
    together. Note that the :class:`.DepressionFinderAndRouter`
    is not particularly intelligent when running on lossy streams, and in particular,
    it will reroute flow around pits even when they are in fact not filled due to loss.

    .. note::

        The perimeter nodes *NEVER* contribute to the accumulating flux, even
        if the  gradients from them point inwards to the main body of the grid.
        This is because under Landlab definitions, perimeter nodes lack cells, so
        cannot accumulate any discharge.

    :class:`~.LossyFlowAccumulator` stores as :class:`~.ModelGrid` fields:

    -  Node array of drainage areas: ``"drainage_area"``
    -  Node array of discharges: ``"surface_water__discharge"``
    -  Node array of discharge loss in transit (vol / sec). This is the
       total loss across all of the downstream links:
       ``"surface_water__discharge_loss"``
    -  Node array containing downstream-to-upstream ordered list of node
       IDs: ``"flow__upstream_node_order"``
    -  Node array of all but the first element of the delta data structure:
       ``"flow__data_structure_delta"``. The first element is always zero.

    The :class:`FlowDirector` component will add additional
    :class:`~.ModelGrid` fields; see the
    :class:`~.FlowAccumulator` for full details. These are:

    -  Node array of receivers (nodes that receive flow), or *ITS OWN ID* if
       there is no receiver: ``"flow__receiver_node"``.
    -  Node array of flow proportions: ``"flow__receiver_proportions"``.
    -  Node array of links carrying flow:  ``"flow__link_to_receiver_node"``.
    -  Node array of downhill slopes from each receiver:
       ``"topographic__steepest_slope"``.
    -  Boolean node array of all local lows: ``"flow__sink_flag"``.

    The primary method of this class is
    :meth:`~landlab.components.LossyFlowAccumulator.run_one_step`.

    Examples
    --------
    These examples pertain only to the :class:`~.LossyFlowAccumulator`.
    See the main :class:`~.FlowAccumulator` documentation for more
    generic and comprehensive examples.

    First, a very simple example. Here's a 50% loss of discharge every time
    flow moves along a node:

    >>> import numpy as np
    >>> from landlab import RasterModelGrid, HexModelGrid
    >>> from landlab.components import FlowDirectorSteepest
    >>> from landlab.components import DepressionFinderAndRouter

    >>> mg = RasterModelGrid((3, 5), xy_spacing=(2, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, False, True)
    >>> z = mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")

    >>> def mylossfunction(qw):
    ...     return 0.5 * qw

    >>> fa = LossyFlowAccumulator(
    ...     mg,
    ...     "topographic__elevation",
    ...     flow_director=FlowDirectorSteepest,
    ...     loss_function=mylossfunction,
    ... )
    >>> fa.run_one_step()

    >>> mg.at_node["drainage_area"].reshape(mg.shape)
    array([[ 0.,  0.,  0.,  0.,  0.],
           [ 6.,  6.,  4.,  2.,  0.],
           [ 0.,  0.,  0.,  0.,  0.]])
    >>> mg.at_node["surface_water__discharge"].reshape(mg.shape)
    array([[ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ],
           [ 1.75,  3.5 ,  3.  ,  2.  ,  0.  ],
           [ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ]])
    >>> mg.at_node["surface_water__discharge_loss"].reshape(mg.shape)
    array([[ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ],
           [ 0.  ,  1.75,  1.5 ,  1.  ,  0.  ],
           [ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ]])

    Here we use a spatially distributed field to derive loss terms, and also
    use a filled, non-raster grid.

    >>> dx = (2.0 / (3.0**0.5)) ** 0.5  # area to be 100.0
    >>> hmg = HexModelGrid((5, 3), spacing=dx, xy_of_lower_left=(-1.0745, 0.0))
    >>> z = hmg.add_field(
    ...     "topographic__elevation", hmg.node_x**2 + np.round(hmg.node_y)**2, at="node"
    ... )
    >>> z[9] = -10.0  # poke a hole
    >>> lossy = hmg.add_zeros("mylossterm", dtype=float, at="node")
    >>> lossy[14] = 1.0  # suppress all flow from node 14

    Without loss looks like this:

    >>> fa = LossyFlowAccumulator(
    ...     hmg,
    ...     "topographic__elevation",
    ...     flow_director=FlowDirectorSteepest,
    ...     depression_finder=DepressionFinderAndRouter,
    ... )
    >>> fa.run_one_step()
    >>> hmg.at_node["flow__receiver_node"]
    array([ 0,  1,  2,
            3,  0,  9,  6,
            7,  9,  4,  9, 11,
           12,  9,  9, 15,
           16, 17, 18])
    >>> np.round(hmg.at_node["drainage_area"])
    array([ 7.,  0.,  0.,
            0.,  7.,  1.,  0.,
            0.,  1.,  6.,  1.,  0.,
            0., 1.,  1.,  0.,
            0.,  0.,  0.])
    >>> np.round(hmg.at_node["surface_water__discharge"])
    array([ 7.,  0.,  0.,
            0.,  7.,  1.,  0.,
            0.,  1.,  6.,  1.,  0.,
            0., 1.,  1.,  0.,
            0.,  0.,  0.])

    With loss looks like this:

    >>> def mylossfunction2(Qw, nodeID, linkID, grid):
    ...     return (1. - grid.at_node['mylossterm'][nodeID]) * Qw
    >>> fa = LossyFlowAccumulator(
    ...     hmg,
    ...     "topographic__elevation",
    ...     flow_director=FlowDirectorSteepest,
    ...     depression_finder=DepressionFinderAndRouter,
    ...     loss_function=mylossfunction2,
    ... )
    >>> fa.run_one_step()
    >>> np.round(hmg.at_node["drainage_area"])
    array([ 7.,  0.,  0.,
            0.,  7.,  1.,  0.,
            0.,  1.,  6.,  1.,  0.,
            0., 1.,  1.,  0.,
            0.,  0.,  0.])
    >>> np.round(hmg.at_node["surface_water__discharge"])
    array([ 6.,  0.,  0.,
            0.,  6.,  1.,  0.,
            0.,  1.,  5.,  1.,  0.,
            0., 1.,  1.,  0.,
            0.,  0.,  0.])
    >>> np.allclose(
    ...     hmg.at_node["surface_water__discharge_loss"],
    ...     lossy * hmg.at_node["surface_water__discharge"],
    ... )
    True

    (Loss is only happening from the node, 14, that we set it to happen at.)

    Finally, note we can use the *linkIDs* to create flow-length-dependent
    effects:

    >>> from landlab.components import FlowDirectorMFD
    >>> mg = RasterModelGrid((4, 6), xy_spacing=(1, 2))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, False, True)
    >>> z = mg.add_field("topographic__elevation", 2.0 * mg.node_x, at="node")
    >>> z[9] = 8.
    >>> z[16] = 6.5  # force the first node sideways

    >>> L = mg.add_zeros("spatialloss", at="node")
    >>> mg.at_node["spatialloss"][9] = 1.
    >>> mg.at_node["spatialloss"][13] = 1.
    >>> def fancyloss(Qw, nodeID, linkID, grid):
    ...     # now a true transmission loss:
    ...     Lt = (1.0 - 1.0 / grid.length_of_link[linkID] ** 2)
    ...     Lsp = grid.at_node["spatialloss"][nodeID]
    ...     return Qw * (1.0 - Lt) * (1.0 - Lsp)

    >>> fa = LossyFlowAccumulator(
    ...     mg,
    ...     "topographic__elevation",
    ...     flow_director=FlowDirectorMFD,
    ...     loss_function=fancyloss,
    ... )
    >>> fa.run_one_step()

    >>> mg.at_node["drainage_area"].reshape(mg.shape)
    array([[  0. ,   0. ,   0. ,   0. ,   0. ,   0. ],
           [  5.6,   5.6,   3.6,   2. ,   2. ,   0. ],
           [ 10.4,  10.4,   8.4,   6.4,   4. ,   0. ],
           [  0. ,   0. ,   0. ,   0. ,   0. ,   0. ]])
    >>> mg.at_node["surface_water__discharge"].reshape(mg.shape)
    array([[ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 4. ,  4. ,  2. ,  2. ,  2. ,  0. ],
           [ 0. ,  8.5,  6.5,  4.5,  2.5,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ]])

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    Braun, J., Willett, S. (2013). A very efficient O(n), implicit and parallel
    method to solve the stream power equation governing fluvial incision and
    landscape evolution. Geomorphology  180-181(C), 170-179.
    https://dx.doi.org/10.1016/j.geomorph.2012.10.008
    """

    _name = "LossyFlowAccumulator"

    _info = {
        "drainage_area": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m**2",
            "mapping": "node",
            "doc": "Upstream accumulated surface area contributing to the node's discharge",
        },
        "flow__data_structure_delta": {
            "dtype": int,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": (
                "Node array containing the elements delta[1:] of the data "
                "structure 'delta' used for construction of the "
                "downstream-to-upstream node array"
            ),
        },
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "surface_water__discharge": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m**3/s",
            "mapping": "node",
            "doc": "Volumetric discharge of surface water",
        },
        "surface_water__discharge_loss": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m**3/s",
            "mapping": "node",
            "doc": "Total volume of water per second lost during all flow out of the node",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "water__unit_flux_in": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "m/s",
            "mapping": "node",
            "doc": (
                "External volume water per area per time input to each node "
                "(e.g., rainfall rate)"
            ),
        },
    }

    def __init__(
        self,
        grid,
        surface="topographic__elevation",
        flow_director="FlowDirectorSteepest",
        runoff_rate=None,
        depression_finder=None,
        loss_function=None,
        **kwargs,
    ):
        """Initialize the FlowAccumulator component.

        Saves the grid, tests grid type, tests imput types and
        compatability for the flow_director and depression_finder
        keyword arguments, tests the argument of runoff_rate, and
        initializes new fields.

        Parameters
        ----------
        grid : ModelGrid
            A Landlab grid.
        surface : str, int or array of int
            The surface to direct flow across.  Can be a field name at node or
            an array of length *node*.
        flow_director : str, class or instance of class.
            A string of method or class name (e.g. ``'D8'`` or ``'FlowDirectorD8'``), an
            uninstantiated FlowDirector class, or an instance of a
            :class:`FlowDirector` class. This sets the method
            used to calculate flow directions.
        runoff_rate : field name, array, or float, optional (m/time)
            If provided, sets the runoff rate and will be assigned to the grid
            field ``'water__unit_flux_in'``. If a spatially and and temporally variable
            runoff rate is desired, pass this field name and update the field
            through model run time. If both the field and argument are present at
            the time of initialization, runoff_rate will *overwrite* the field. If
            neither are set, defaults to spatially constant unit input.
        depression_finder : str, class, instance of class, optional
            A string of class name (e.g., ``'DepressionFinderAndRouter'``), an
            uninstantiated :class:`~.DepressionFinder` class,
            or an instance of a :class:`~.DepressionFinder` class.
            This sets the method for depression finding.
        loss_function : func, optional
            A function of the form ``f(Qw, [node_ID, [linkID, [grid]]])``, where Qw is
            the discharge at a node, node_ID the ID of the node at which the loss
            is to be calculated, linkID is the ID of the link down which the
            outflow drains (or a d8 ID if the routing is d8), and grid is a Landlab
            ModelGrid. The function then returns the new discharge at the node
            after the function is applied.

            Note that if a linkID is needed, a nodeID must also be specified, even
            if only as a dummy parameter; similarly, if a grid is to be passed, all
            of the preceding parameters must be specified. Both nodeID and linkID
            are required to permit spatially variable losses, and also losses
            dependent on flow path geometry (e.g., flow length). The grid is passed
            to allow fields or grid properties describing values across the grid
            to be accessed for the loss calculation (see examples).
            This function expects (float, [int, [int, [ModelGrid]]]), and
            return a single float, the new discharge value. This behavior is
            verified during component instantiation.
        **kwargs : optional
            Any additional parameters to pass to a FlowDirector or
            DepressionFinderAndRouter instance (e.g., partion_method for
            FlowDirectorMFD). This will have no effect if an instantiated
            component is passed using the flow_director or depression_finder
            keywords.
        """

        # add the new loss discharge field if necessary:
        if "surface_water__discharge_loss" not in grid.at_node:
            grid.add_zeros(
                "node", "surface_water__discharge_loss", dtype=float, clobber=True
            )

        super().__init__(
            grid,
            surface=surface,
            flow_director=flow_director,
            runoff_rate=runoff_rate,
            depression_finder=depression_finder,
            **kwargs,
        )

        if loss_function is not None:
            if sys.version_info[0] >= 3:
                sig = signature(loss_function)
                num_params = len(sig.parameters)
            else:  # Python 2
                num_params = loss_function.func_code.co_argcount
            # save the func for loss, and do a quick test on its inputs:
            if num_params == 1:
                # check the func takes a single value and turns it into a new
                # single value:
                if not isinstance(loss_function(1.0), float):
                    raise TypeError(
                        "The loss_function should take a float, and return " "a float."
                    )
                # now, for logical consistency in our calls to
                # find_drainage_area_and_discharge, wrap the func so it has two
                # arguments:

                def lossfunc(Qw, dummyn, dummyl, dummygrid):
                    return float(loss_function(Qw))

                self._lossfunc = lossfunc

            elif num_params == 2:
                # check the func takes a single value and turns it into a new
                # single value:
                if not isinstance(loss_function(1.0, 0), float):
                    raise TypeError(
                        "The loss_function should take (float, int), and "
                        "return a float."
                    )
                # now, for logical consistency in our calls to
                # find_drainage_area_and_discharge, wrap the func so it has two
                # arguments:

                def lossfunc(Qw, nodeID, dummyl, dummygrid):
                    return float(loss_function(Qw, nodeID))

                self._lossfunc = lossfunc

            elif num_params == 3:
                # check the func takes (float, int) and turns it into a new
                # single value:
                if not isinstance(loss_function(1.0, 0, 0), float):
                    raise TypeError(
                        "The loss_function should take (float, int, int), "
                        "and return a float."
                    )

                def lossfunc(Qw, nodeID, linkID, dummygrid):
                    return float(loss_function(Qw, nodeID, linkID))

                self._lossfunc = lossfunc

            elif num_params == 4:
                # this time, the test is too hard to implement cleanly so just
                self._lossfunc = loss_function
            else:
                raise ValueError(
                    "The loss_function must have only a single argument, "
                    "which should be the discharge at a node; a pair of "
                    "arguments, which should be the discharge at a node and "
                    "the node ID; or three arguments, which should be the "
                    "discharge at a node, the node ID, and the link along "
                    "which that discharge will flow."
                )
        else:
            # make a dummy
            def lossfunc(Qw, dummyn, dummyl, dummygrid):
                return float(Qw)

            self._lossfunc = lossfunc

    def _accumulate_A_Q_to_one(self, s, r):
        """Accumulate area and discharge for a route-to-one scheme."""
        link = self._grid.at_node["flow__link_to_receiver_node"]
        a, q = flow_accum_bw.find_drainage_area_and_discharge_lossy(
            s,
            r,
            link,
            self._lossfunc,
            self._grid,
            self._node_cell_area,
            self._grid.at_node["water__unit_flux_in"],
        )
        return a, q

    def _accumulate_A_Q_to_n(self, s, r, p):
        """Accumulate area and discharge for a route-to-one scheme."""
        link = self._grid.at_node["flow__link_to_receiver_node"]
        a, q = flow_accum_to_n.find_drainage_area_and_discharge_to_n_lossy(
            s,
            r,
            link,
            p,
            self._lossfunc,
            self._grid,
            self._node_cell_area,
            self._grid.at_node["water__unit_flux_in"],
        )
        return a, q
